import contextlib
import enum
from dataclasses import dataclass
from typing import Any, Optional

from vyper.codegen.ir_node import Encoding, IRnode
from vyper.compiler.settings import get_global_settings
from vyper.evm.address_space import MEMORY, AddrSpace
from vyper.exceptions import CompilerPanic, StateAccessViolation
from vyper.semantics.types import VyperType


class Constancy(enum.Enum):
    Mutable = 0
    Constant = 1


@dataclass(frozen=True)
class Alloca:
    name: str
    offset: int
    typ: VyperType
    size: int

    def __post_init__(self):
        assert self.typ.memory_bytes_required == self.size


# Function variable
@dataclass
class VariableRecord:
    name: str
    pos: int
    typ: VyperType
    mutable: bool
    encoding: Encoding = Encoding.VYPER
    location: AddrSpace = MEMORY
    size: Optional[int] = None  # allocated size
    blockscopes: Optional[list] = None
    defined_at: Any = None
    is_internal: bool = False
    alloca: Optional[Alloca] = None

    # the following members are probably dead
    is_immutable: bool = False
    is_transient: bool = False
    data_offset: Optional[int] = None

    def __hash__(self):
        return hash(id(self))

    def __post_init__(self):
        if self.blockscopes is None:
            self.blockscopes = []

    def __repr__(self):
        ret = vars(self)
        ret["allocated"] = self.typ.memory_bytes_required
        return f"VariableRecord({ret})"

    def as_ir_node(self):
        ret = IRnode.from_list(
            self.pos,
            typ=self.typ,
            annotation=self.name,
            encoding=self.encoding,
            mutable=self.mutable,
            location=self.location,
        )
        ret._referenced_variables = {self}
        if self.alloca is not None:
            ret.passthrough_metadata["alloca"] = self.alloca
        return ret


# compilation context for a function
class Context:
    def __init__(
        self,
        module_ctx,
        memory_allocator,
        vars_=None,
        forvars=None,
        constancy=Constancy.Mutable,
        func_t=None,
        is_ctor_context=False,
    ):
        # In-memory variables, in the form (name, memory location, type)
        self.vars = vars_ or {}

        # Variables defined in for loops, e.g. for i in range(6): ...
        self.forvars = forvars or {}

        # Is the function constant?
        self.constancy = constancy

        # Whether we are currently parsing a range expression
        self.in_range_expr = False

        # store module context
        self.module_ctx = module_ctx

        # full function type
        self.func_t = func_t
        # Active scopes
        self._scopes = set()

        # Memory allocator, keeps track of currently allocated memory.
        # Not intended to be accessed directly
        self.memory_allocator = memory_allocator

        # save the starting memory location so we can find out (later)
        # how much memory this function uses.
        self.starting_memory = memory_allocator.next_mem

        # Incremented values, used for internal IDs
        self._internal_var_iter = 0
        self._scope_id_iter = 0

        # either the constructor, or called from the constructor
        self.is_ctor_context = is_ctor_context

        self.settings = get_global_settings()

    def is_constant(self):
        return self.constancy is Constancy.Constant or self.in_range_expr

    def check_is_not_constant(self, err, expr):
        if self.is_constant():
            raise StateAccessViolation(f"Cannot {err} from {self.pp_constancy()}", expr)

    # convenience properties
    @property
    def is_payable(self):
        return self.func_t.is_payable

    @property
    def is_internal(self):
        return self.func_t.is_internal

    @property
    def return_type(self):
        return self.func_t.return_type

    #
    # Context Managers
    # - Context managers are used to ensure proper wrapping of scopes and context states.

    @contextlib.contextmanager
    def range_scope(self):
        prev_value = self.in_range_expr
        self.in_range_expr = True
        yield
        self.in_range_expr = prev_value

    @contextlib.contextmanager
    def internal_memory_scope(self):
        """
        Internal memory scope context manager.

        Internal variables that are declared within this context are de-allocated
        upon exiting the context.
        """
        scope_id = self._scope_id_iter
        self._scope_id_iter += 1
        self._scopes.add(scope_id)
        yield

        # Remove all variables that have specific scope_id attached
        released = [
            (k, v) for k, v in self.vars.items() if v.is_internal and scope_id in v.blockscopes
        ]
        for name, var in released:
            self.deallocate_variable(name, var)

        # Remove block scopes
        self._scopes.remove(scope_id)

    @contextlib.contextmanager
    def block_scope(self):
        """
        Block scope context manager.

        All variables (public and internal) that are declared within this context
        are de-allocated upon exiting the context.
        """
        scope_id = self._scope_id_iter
        self._scope_id_iter += 1
        self._scopes.add(scope_id)
        yield

        # Remove all variables that have specific scope_id attached
        released = [(k, v) for k, v in self.vars.items() if scope_id in v.blockscopes]
        for name, var in released:
            self.deallocate_variable(name, var)

        # Remove block scopes
        self._scopes.remove(scope_id)

    def deallocate_variable(self, varname, var):
        assert varname == var.name

        # sanity check the type's size hasn't changed since allocation.
        n = var.typ.memory_bytes_required
        assert n == var.size

        if self.settings.experimental_codegen:
            # do not deallocate at this stage because this will break
            # analysis in venom; venom will do its own alloc/dealloc/analysis.
            pass
        else:
            self.memory_allocator.deallocate_memory(var.pos, var.size)

        del self.vars[var.name]

    def _new_variable(
        self,
        name: str,
        typ: VyperType,
        is_internal: bool,
        is_mutable: bool = True,
        internal_function=False,
    ) -> IRnode:
        size = typ.memory_bytes_required

        ofst = self.memory_allocator.allocate_memory(size)
        assert ofst + size <= self.memory_allocator.size_of_mem, "function frame overrun"

        pos = ofst
        alloca = None
        if self.settings.experimental_codegen:
            # convert it into an abstract pointer
            if internal_function:
                pos = f"$palloca_{ofst}_{size}"
            else:
                pos = f"$alloca_{ofst}_{size}"
            alloca = Alloca(name=name, offset=ofst, typ=typ, size=size)

        var = VariableRecord(
            name=name,
            pos=pos,
            typ=typ,
            size=size,
            mutable=is_mutable,
            blockscopes=self._scopes.copy(),
            is_internal=is_internal,
            alloca=alloca,
        )
        self.vars[name] = var
        return var.as_ir_node()

    def new_variable(
        self, name: str, typ: VyperType, is_mutable: bool = True, internal_function=False
    ) -> IRnode:
        """
        Allocate memory for a user-defined variable and return an IR node referencing it.

        Arguments
        ---------
        name : str
            Name of the variable
        typ : VyperType
            Variable type, used to determine the size of memory allocation

        Returns
        -------
        int
            Memory offset for the variable
        """

        return self._new_variable(
            name, typ, is_internal=False, is_mutable=is_mutable, internal_function=internal_function
        )

    def fresh_varname(self, name: str) -> str:
        """
        return a unique variable name
        """
        t = self._internal_var_iter
        self._internal_var_iter += 1
        return f"{name}{t}"

    def new_internal_variable(self, typ: VyperType) -> IRnode:
        """
        Allocate memory for an internal variable.

        Arguments
        ---------
        typ : VyperType
            Variable type, used to determine the size of memory allocation

        Returns
        -------
        int
            Memory offset for the variable
        """
        # internal variable names begin with a number sign so there is no chance for collision
        name = self.fresh_varname("#internal")

        return self._new_variable(name, typ, is_internal=True)

    def lookup_var(self, varname) -> VariableRecord:
        return self.vars[varname]

    # Pretty print constancy for error messages
    def pp_constancy(self):
        if self.in_range_expr:
            return "a range expression"
        elif self.constancy == Constancy.Constant:
            return "a constant function"
        raise CompilerPanic(f"bad constancy: {self.constancy}")  # pragma: nocover



def test_export_nonreentrant(make_input_bundle, get_contract, tx_failed):
    lib1 = """
interface Foo:
    def foo() -> uint256: nonpayable

implements: Foo

@external
@nonreentrant
def foo() -> uint256:
    return 5
    """
    main = """
import lib1

initializes: lib1

exports: lib1.foo

@external
@nonreentrant
def re_enter():
    extcall lib1.Foo(self).foo()  # should always throw

@external
def __default__():
    # sanity: make sure we don't revert due to bad selector
    pass
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 5
    with tx_failed():
        c.re_enter()


def test_internal_nonreentrant(make_input_bundle, get_contract, tx_failed):
    lib1 = """
interface Foo:
    def foo() -> uint256: nonpayable

implements: Foo

@external
def foo() -> uint256:
    return self._safe_fn()

@internal
@nonreentrant
def _safe_fn() -> uint256:
    return 10
    """
    main = """
import lib1

initializes: lib1

exports: lib1.foo

@external
@nonreentrant
def re_enter():
    extcall lib1.Foo(self).foo()  # should always throw

@external
def __default__():
    # sanity: make sure we don't revert due to bad selector
    pass
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 10
    with tx_failed():
        c.re_enter()



import pytest
from eth.codecs import abi

from tests.utils import decimal_to_int


# @pytest.mark.parametrize("string", ["a", "abc", "abcde", "potato"])
def test_abi_encode(get_contract):
    code = """
struct Animal:
  name: String[5]
  address_: address
  id_: int128
  is_furry: bool
  price: decimal
  data: uint256[3]
  metadata: bytes32

struct Human:
  name: String[64]
  pet: Animal

@external
# TODO accept struct input once the functionality is available
def abi_encode(
    name: String[64],
    pet_name: String[5],
    pet_address: address,
    pet_id: int128,
    pet_is_furry: bool,
    pet_price: decimal,
    pet_data: uint256[3],
    pet_metadata: bytes32,
    ensure_tuple: bool,
    include_method_id: bool
) -> Bytes[548]:
    human: Human = Human(
      name=name,
      pet=Animal(
        name=pet_name,
        address_=pet_address,
        id_=pet_id,
        is_furry=pet_is_furry,
        price=pet_price,
        data=pet_data,
        metadata=pet_metadata
      ),
    )
    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(human) # default ensure_tuple=True
        return _abi_encode(human, method_id=0xdeadbeef)
    else:
        if not include_method_id:
            return _abi_encode(human, ensure_tuple=False)
        return _abi_encode(human, ensure_tuple=False, method_id=0xdeadbeef)

@external
def abi_encode2(name: String[32], ensure_tuple: bool, include_method_id: bool) -> Bytes[100]:
    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(name) # default ensure_tuple=True
        return _abi_encode(name, method_id=0xdeadbeef)
    else:
        if not include_method_id:
            return _abi_encode(name, ensure_tuple=False)
        return _abi_encode(name, ensure_tuple=False, method_id=0xdeadbeef)

@external
def abi_encode3(x: uint256, ensure_tuple: bool, include_method_id: bool) -> Bytes[36]:

    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(x) # default ensure_tuple=True

        return _abi_encode(x, method_id=0xdeadbeef)

    else:
        if not include_method_id:
            return _abi_encode(x, ensure_tuple=False)

        return _abi_encode(x, ensure_tuple=False, method_id=0xdeadbeef)
    """
    c = get_contract(code)

    method_id = 0xDEADBEEF.to_bytes(4, "big")

    # test each method once each with ensure_tuple set to True and False

    arg = 123
    assert c.abi_encode3(arg, False, False).hex() == abi.encode("uint256", arg).hex()
    assert c.abi_encode3(arg, True, False).hex() == abi.encode("(uint256)", (arg,)).hex()
    assert c.abi_encode3(arg, False, True).hex() == (method_id + abi.encode("uint256", arg)).hex()
    assert (
        c.abi_encode3(arg, True, True).hex() == (method_id + abi.encode("(uint256)", (arg,))).hex()
    )

    arg = "some string"
    assert c.abi_encode2(arg, False, False).hex() == abi.encode("string", arg).hex()
    assert c.abi_encode2(arg, True, False).hex() == abi.encode("(string)", (arg,)).hex()
    assert c.abi_encode2(arg, False, True).hex() == (method_id + abi.encode("string", arg)).hex()
    assert (
        c.abi_encode2(arg, True, True).hex() == (method_id + abi.encode("(string)", (arg,))).hex()
    )

    test_addr = "0x" + b"".join(chr(i).encode("utf-8") for i in range(20)).hex()
    test_bytes32 = b"".join(chr(i).encode("utf-8") for i in range(32))
    human_tuple = (
        "foobar",
        ("vyper", test_addr, 123, True, decimal_to_int("123.4"), [123, 456, 789], test_bytes32),
    )
    args = tuple([human_tuple[0]] + list(human_tuple[1]))
    human_t = "(string,(string,address,int128,bool,int168,uint256[3],bytes32))"
    human_encoded = abi.encode(human_t, human_tuple)
    assert c.abi_encode(*args, False, False).hex() == human_encoded.hex()
    assert c.abi_encode(*args, False, True).hex() == (method_id + human_encoded).hex()

    human_encoded = abi.encode(f"({human_t})", (human_tuple,))
    assert c.abi_encode(*args, True, False).hex() == human_encoded.hex()
    assert c.abi_encode(*args, True, True).hex() == (method_id + human_encoded).hex()


@pytest.mark.parametrize("type,value", [("Bytes", b"hello"), ("String", "hello")])
def test_abi_encode_length_failing(get_contract, assert_compile_failed, type, value):
    code = f"""
struct WrappedBytes:
    bs: {type}[6]

@internal
def foo():
    x: WrappedBytes = WrappedBytes(bs={value})
    y: {type}[96] = _abi_encode(x, ensure_tuple=True) # should be Bytes[128]
    """

    assert_compile_failed(lambda: get_contract(code))


def test_abi_encode_dynarray(get_contract):
    code = """
@external
def abi_encode(d: DynArray[uint256, 3], ensure_tuple: bool, include_method_id: bool) -> Bytes[164]:
    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(d) # default ensure_tuple=True
        return _abi_encode(d, method_id=0xdeadbeef)
    else:
        if not include_method_id:
            return _abi_encode(d, ensure_tuple=False)
        return _abi_encode(d, ensure_tuple=False, method_id=0xdeadbeef)
    """
    c = get_contract(code)

    method_id = 0xDEADBEEF.to_bytes(4, "big")

    arg = [123, 456, 789]
    assert c.abi_encode(arg, False, False).hex() == abi.encode("uint256[]", arg).hex()
    assert c.abi_encode(arg, True, False).hex() == abi.encode("(uint256[])", (arg,)).hex()
    assert c.abi_encode(arg, False, True).hex() == (method_id + abi.encode("uint256[]", arg)).hex()
    assert (
        c.abi_encode(arg, True, True).hex() == (method_id + abi.encode("(uint256[])", (arg,))).hex()
    )


nested_2d_array_args = [
    [[123, 456, 789], [234, 567, 891], [345, 678, 912]],
    [[], [], []],
    [[123, 456], [234, 567, 891]],
    [[123, 456, 789], [234, 567], [345]],
    [[123], [], [345, 678, 912]],
    [[], [], [345, 678, 912]],
    [[], [], [345]],
    [[], [234], []],
    [[], [234, 567, 891], []],
    [[]],
    [[123], [234]],
]


@pytest.mark.parametrize("args", nested_2d_array_args)
def test_abi_encode_nested_dynarray(get_contract, args):
    code = """
@external
def abi_encode(
    d: DynArray[DynArray[uint256, 3], 3], ensure_tuple: bool, include_method_id: bool
) -> Bytes[548]:
    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(d) # default ensure_tuple=True
        return _abi_encode(d, method_id=0xdeadbeef)
    else:
        if not include_method_id:
            return _abi_encode(d, ensure_tuple=False)
        return _abi_encode(d, ensure_tuple=False, method_id=0xdeadbeef)
    """
    c = get_contract(code)

    method_id = 0xDEADBEEF.to_bytes(4, "big")

    assert c.abi_encode(args, False, False).hex() == abi.encode("uint256[][]", args).hex()
    assert c.abi_encode(args, True, False).hex() == abi.encode("(uint256[][])", (args,)).hex()
    assert (
        c.abi_encode(args, False, True).hex() == (method_id + abi.encode("uint256[][]", args)).hex()
    )
    assert (
        c.abi_encode(args, True, True).hex()
        == (method_id + abi.encode("(uint256[][])", (args,))).hex()
    )


nested_3d_array_args = [
    [
        [[123, 456, 789], [234, 567, 891], [345, 678, 912]],
        [[234, 567, 891], [345, 678, 912], [123, 456, 789]],
        [[345, 678, 912], [123, 456, 789], [234, 567, 891]],
    ],
    [[[123, 789], [234], [345, 678, 912]], [[234, 567], [345, 678]], [[345]]],
    [[[123], [234, 567, 891]], [[234]]],
    [[[], [], []], [[], [], []], [[], [], []]],
    [[[123, 456, 789], [234, 567, 891], [345, 678, 912]], [[234, 567, 891], [345, 678, 912]], [[]]],
    [[[]], [[]], [[234]]],
    [[[123]], [[]], [[]]],
    [[[]], [[123]], [[]]],
    [[[123, 456, 789], [234, 567]], [[234]], [[567], [912], [345]]],
    [[[]]],
]


@pytest.mark.parametrize("args", nested_3d_array_args)
def test_abi_encode_nested_dynarray_2(get_contract, args):
    code = """
@external
def abi_encode(
    d: DynArray[DynArray[DynArray[uint256, 3], 3], 3],
    ensure_tuple: bool,
    include_method_id: bool
) -> Bytes[1700]:
    if ensure_tuple:
        if not include_method_id:
            return _abi_encode(d) # default ensure_tuple=True
        return _abi_encode(d, method_id=0xdeadbeef)
    else:
        if not include_method_id:
            return _abi_encode(d, ensure_tuple=False)
        return _abi_encode(d, ensure_tuple=False, method_id=0xdeadbeef)
    """
    c = get_contract(code)

    method_id = 0xDEADBEEF.to_bytes(4, "big")

    assert c.abi_encode(args, False, False).hex() == abi.encode("uint256[][][]", args).hex()
    assert c.abi_encode(args, True, False).hex() == abi.encode("(uint256[][][])", (args,)).hex()
    assert (
        c.abi_encode(args, False, True).hex()
        == (method_id + abi.encode("uint256[][][]", args)).hex()
    )
    assert (
        c.abi_encode(args, True, True).hex()
        == (method_id + abi.encode("(uint256[][][])", (args,))).hex()
    )


def test_side_effects_evaluation(get_contract):
    contract_1 = """
counter: uint256

@deploy
def __init__():
    self.counter = 0

@external
def get_counter() -> (uint256, String[6]):
    self.counter += 1
    return (self.counter, "hello")
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def get_counter() -> (uint256, String[6]): nonpayable

@external
def foo(addr: address) -> Bytes[164]:
    return _abi_encode(extcall Foo(addr).get_counter(), method_id=0xdeadbeef)
    """

    c2 = get_contract(contract_2)

    method_id = 0xDEADBEEF.to_bytes(4, "big")

    # call to get_counter() should be evaluated only once
    get_counter_encoded = abi.encode("((uint256,string))", ((1, "hello"),))

    assert c2.foo(c.address).hex() == (method_id + get_counter_encoded).hex()


# test _abi_encode in private functions to check buffer overruns
def test_abi_encode_private(get_contract):
    code = """
bytez: Bytes[96]
@internal
def _foo(bs: Bytes[32]):
    self.bytez = _abi_encode(bs)

@external
def foo(bs: Bytes[32]) -> (uint256, Bytes[96]):
    dont_clobber_me: uint256 = max_value(uint256)
    self._foo(bs)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    bs = b"\x00" * 32
    assert c.foo(bs) == (2**256 - 1, abi.encode("(bytes)", (bs,)))


def test_abi_encode_private_dynarray(get_contract):
    code = """
bytez: Bytes[160]
@internal
def _foo(bs: DynArray[uint256, 3]):
    self.bytez = _abi_encode(bs)
@external
def foo(bs: DynArray[uint256, 3]) -> (uint256, Bytes[160]):
    dont_clobber_me: uint256 = max_value(uint256)
    self._foo(bs)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    bs = [1, 2, 3]
    assert c.foo(bs) == (2**256 - 1, abi.encode("(uint256[])", (bs,)))


def test_abi_encode_private_nested_dynarray(get_contract):
    code = """
bytez: Bytes[1696]
@internal
def _foo(bs: DynArray[DynArray[DynArray[uint256, 3], 3], 3]):
    self.bytez = _abi_encode(bs)

@external
def foo(bs: DynArray[DynArray[DynArray[uint256, 3], 3], 3]) -> (uint256, Bytes[1696]):
    dont_clobber_me: uint256 = max_value(uint256)
    self._foo(bs)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    bs = [
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        [[10, 11, 12], [13, 14, 15], [16, 17, 18]],
        [[19, 20, 21], [22, 23, 24], [25, 26, 27]],
    ]
    assert c.foo(bs) == (2**256 - 1, abi.encode("(uint256[][][])", (bs,)))


@pytest.mark.parametrize("empty_literal", ('b""', '""', "empty(Bytes[1])", "empty(String[1])"))
def test_abi_encode_empty_string(get_contract, empty_literal):
    code = f"""
@external
def foo(ensure_tuple: bool) -> Bytes[96]:
    if ensure_tuple:
        return _abi_encode({empty_literal}) # default ensure_tuple=True
    else:
        return _abi_encode({empty_literal}, ensure_tuple=False)
    """

    c = get_contract(code)

    # eth-abi does not encode zero-length string correctly -
    # see https://github.com/ethereum/eth-abi/issues/157
    expected_output = b"\x00" * 32
    assert c.foo(False) == expected_output
    expected_output = b"\x00" * 31 + b"\x20" + b"\x00" * 32
    assert c.foo(True) == expected_output



import pytest

import vyper
from vyper.compiler.output import _build_opcodes
from vyper.evm import opcodes
from vyper.exceptions import CompilerPanic


def test_opcodes():
    code = """
@external
def a() -> bool:
    return True
    """

    out = vyper.compile_code(code, output_formats=["opcodes_runtime", "opcodes"])

    assert len(out["opcodes"]) > len(out["opcodes_runtime"])
    assert out["opcodes_runtime"] in out["opcodes"]


def test_version_check_no_begin_or_end():
    with pytest.raises(CompilerPanic):
        opcodes.version_check()


def test_version_check(evm_version):
    assert opcodes.version_check(begin=evm_version)
    assert opcodes.version_check(end=evm_version)
    assert opcodes.version_check(begin=evm_version, end=evm_version)
    if evm_version not in ("london",):
        assert not opcodes.version_check(end="london")
    london_check = opcodes.version_check(begin="london")
    assert london_check == (opcodes.EVM_VERSIONS[evm_version] >= opcodes.EVM_VERSIONS["london"])


def test_get_opcodes(evm_version):
    ops = opcodes.get_opcodes()

    assert "CHAINID" in ops
    assert ops["CREATE2"][-1] == 32000

    assert ops["SLOAD"][-1] == 2100

    if evm_version in ("shanghai", "cancun"):
        assert "PUSH0" in ops

    if evm_version in ("cancun",):
        for op in ("TLOAD", "TSTORE", "MCOPY"):
            assert op in ops
    else:
        for op in ("TLOAD", "TSTORE", "MCOPY"):
            assert op not in ops


def test_build_opcodes():
    assert _build_opcodes(bytes.fromhex("610250")) == "PUSH2 0x0250"
    assert _build_opcodes(bytes.fromhex("612500")) == "PUSH2 0x2500"
    assert _build_opcodes(bytes.fromhex("610100")) == "PUSH2 0x0100"
    assert _build_opcodes(bytes.fromhex("611000")) == "PUSH2 0x1000"
    assert _build_opcodes(bytes.fromhex("62010300")) == "PUSH3 0x010300"
    assert (
        _build_opcodes(
            bytes.fromhex("7f6100000000000000000000000000000000000000000000000000000000000000")
        )
        == "PUSH32 0x6100000000000000000000000000000000000000000000000000000000000000"
    )



import itertools

import pytest


def test_short_circuit_and_left_is_false(get_contract):
    code = """

called_left: public(bool)
called_right: public(bool)

@internal
def left() -> bool:
    self.called_left = True
    return False

@internal
def right() -> bool:
    self.called_right = True
    return False

@external
def foo() -> bool:
    return self.left() and self.right()
"""
    c = get_contract(code)
    assert not c.foo()

    c.foo()
    assert c.called_left()
    assert not c.called_right()


def test_short_circuit_and_left_is_true(get_contract):
    code = """

called_left: public(bool)
called_right: public(bool)

@internal
def left() -> bool:
    self.called_left = True
    return True

@internal
def right() -> bool:
    self.called_right = True
    return True

@external
def foo() -> bool:
    return self.left() and self.right()
"""
    c = get_contract(code)
    assert c.foo()

    c.foo()
    assert c.called_left()
    assert c.called_right()


def test_short_circuit_or_left_is_true(get_contract):
    code = """

called_left: public(bool)
called_right: public(bool)

@internal
def left() -> bool:
    self.called_left = True
    return True

@internal
def right() -> bool:
    self.called_right = True
    return True

@external
def foo() -> bool:
    return self.left() or self.right()
"""
    c = get_contract(code)
    assert c.foo()

    c.foo()
    assert c.called_left()
    assert not c.called_right()


def test_short_circuit_or_left_is_false(get_contract):
    code = """

called_left: public(bool)
called_right: public(bool)

@internal
def left() -> bool:
    self.called_left = True
    return False

@internal
def right() -> bool:
    self.called_right = True
    return False

@external
def foo() -> bool:
    return self.left() or self.right()
"""
    c = get_contract(code)
    assert not c.foo()

    c.foo()
    assert c.called_left()
    assert c.called_right()


@pytest.mark.parametrize("op", ["and", "or"])
@pytest.mark.parametrize("a, b", itertools.product([True, False], repeat=2))
def test_from_memory(get_contract, a, b, op):
    code = f"""
@external
def foo(a: bool, b: bool) -> bool:
    c: bool = a
    d: bool = b
    return c {op} d
"""
    c = get_contract(code)
    assert c.foo(a, b) == eval(f"{a} {op} {b}")


@pytest.mark.parametrize("op", ["and", "or"])
@pytest.mark.parametrize("a, b", itertools.product([True, False], repeat=2))
def test_from_storage(get_contract, a, b, op):
    code = f"""
c: bool
d: bool

@external
def foo(a: bool, b: bool) -> bool:
    self.c = a
    self.d = b
    return self.c {op} self.d
"""
    c = get_contract(code)
    assert c.foo(a, b) == eval(f"{a} {op} {b}")


@pytest.mark.parametrize("op", ["and", "or"])
@pytest.mark.parametrize("a, b", itertools.product([True, False], repeat=2))
def test_from_calldata(get_contract, a, b, op):
    code = f"""
@external
def foo(a: bool, b: bool) -> bool:
    return a {op} b
"""
    c = get_contract(code)
    assert c.foo(a, b) == eval(f"{a} {op} {b}")


@pytest.mark.parametrize("a, b, c, d", itertools.product([True, False], repeat=4))
@pytest.mark.parametrize("ops", itertools.product(["and", "or"], repeat=3))
def test_complex_combination(get_contract, a, b, c, d, ops):
    boolop = f"a {ops[0]} b {ops[1]} c {ops[2]} d"

    code = f"""
@external
def foo(a: bool, b: bool, c: bool, d: bool) -> bool:
    return {boolop}
"""
    contract = get_contract(code)
    if eval(boolop):
        assert contract.foo(a, b, c, d)
    else:
        assert not contract.foo(a, b, c, d)



"""
tests for the uses/initializes checker
main properties to test:
- state usage -- if a module uses state, it must `used` or `initialized`
- conversely, if a module does not touch state, it should not be `used`
- global initializer check: each used module is `initialized` exactly once
"""

import pytest

from vyper.compiler import compile_code
from vyper.exceptions import (
    BorrowException,
    ImmutableViolation,
    InitializerException,
    StructureException,
    UndeclaredDefinition,
)

from .helpers import NONREENTRANT_NOTE


def test_initialize_uses(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@deploy
def __init__():
    pass

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib2
import lib1

initializes: lib2[lib1 := lib1]
initializes: lib1

@deploy
def __init__():
    lib1.__init__()
    lib2.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_initialize_multiple_uses(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    lib2 = """
totalSupply: uint256
    """
    lib3 = """
import lib1
import lib2

# multiple uses on one line
uses: (
    lib1,
    lib2
)

counter: uint256

@deploy
def __init__():
    pass

@internal
def foo():
    x: uint256 = lib2.totalSupply
    lib1.counter += 1
    """
    main = """
import lib1
import lib2
import lib3

initializes: lib1
initializes: lib2
initializes: lib3[
    lib1 := lib1,
    lib2 := lib2
]

@deploy
def __init__():
    lib1.__init__()
    lib3.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "lib3.vy": lib3})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_initialize_multi_line_uses(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    lib2 = """
totalSupply: uint256
    """
    lib3 = """
import lib1
import lib2

uses: lib1
uses: lib2

counter: uint256

@deploy
def __init__():
    pass

@internal
def foo():
    x: uint256 = lib2.totalSupply
    lib1.counter += 1
    """
    main = """
import lib1
import lib2
import lib3

initializes: lib1
initializes: lib2
initializes: lib3[
    lib1 := lib1,
    lib2 := lib2
]

@deploy
def __init__():
    lib1.__init__()
    lib3.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "lib3.vy": lib3})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_initialize_uses_attribute(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@deploy
def __init__():
    pass

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
initializes: lib1

@deploy
def __init__():
    lib2.__init__()
    # demonstrate we can call lib1.__init__ through lib2.lib1
    # (not sure this should be allowed, really.
    lib2.lib1.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_initializes_without_init_function(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
initializes: lib1

@deploy
def __init__():
    pass
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_imported_as_different_names(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1 as m

uses: m

counter: uint256

@internal
def foo():
    m.counter += 1
    """
    main = """
import lib1 as some_module
import lib2

initializes: lib2[m := some_module]
initializes: some_module
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_initializer_list_module_mismatch(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
something: uint256
    """
    lib3 = """
import lib1

uses: lib1

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2
import lib3

initializes: lib1
initializes: lib3[lib1 := lib2]  # typo -- should be [lib1 := lib1]
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "lib3.vy": lib3})

    with pytest.raises(StructureException) as e:
        assert compile_code(main, input_bundle=input_bundle) is not None

    assert e.value._message == "lib1 is not lib2!"


def test_imported_as_different_names_error(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1 as m

uses: m

counter: uint256

@internal
def foo():
    m.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(UndeclaredDefinition) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "unknown module `lib1`"
    assert e.value._hint == "did you mean `m := lib1`?"


def test_global_initializer_constraint(make_input_bundle, chdir_tmp_path):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
# forgot to initialize lib1!
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "module `lib1.vy` is used but never initialized!"
    assert e.value._hint == "add `initializes: lib1` to the top level of your main contract"


def test_initializer_no_references(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "`lib2` uses `lib1`, but it is not initialized with `lib1`"
    assert e.value._hint == "did you mean lib2[lib1 := lib1]?"


def test_missing_uses(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

counter: uint256

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib1
import lib2

initializes: lib2
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_read(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

counter: uint256

@internal
def foo() -> uint256:
    return lib1.counter
    """
    main = """
import lib1
import lib2

initializes: lib1

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_read_immutable(make_input_bundle):
    lib1 = """
MY_IMMUTABLE: immutable(uint256)

@deploy
def __init__():
    MY_IMMUTABLE = 7
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

counter: uint256

@internal
def foo() -> uint256:
    return lib1.MY_IMMUTABLE
    """
    main = """
import lib1
import lib2

initializes: lib1

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_read_inside_call(make_input_bundle):
    lib1 = """
MY_IMMUTABLE: immutable(uint256)

@deploy
def __init__():
    MY_IMMUTABLE = 9

@internal
def get_counter() -> uint256:
    return MY_IMMUTABLE
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

counter: uint256

@internal
def foo() -> uint256:
    return lib1.get_counter()
    """
    main = """
import lib1
import lib2

initializes: lib1

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_hashmap(make_input_bundle):
    lib1 = """
counter: HashMap[uint256, HashMap[uint256, uint256]]
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

@internal
def foo() -> uint256:
    return lib1.counter[1][2]
    """
    main = """
import lib1
import lib2

initializes: lib1

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_tuple(make_input_bundle):
    lib1 = """
counter: HashMap[uint256, HashMap[uint256, uint256]]
    """
    lib2 = """
import lib1

interface Foo:
    def foo() -> (uint256, uint256): nonpayable

something: uint256

# forgot `uses: lib1`!

@internal
def foo() -> uint256:
    lib1.counter[1][2], self.something = extcall Foo(msg.sender).foo()
    """
    main = """
import lib1
import lib2

initializes: lib1
initializes: lib2

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_for_tuple_function_call(make_input_bundle):
    lib1 = """
counter: HashMap[uint256, HashMap[uint256, uint256]]

something: uint256

interface Foo:
    def foo() -> (uint256, uint256): nonpayable

@internal
def write_tuple():
    self.counter[1][2], self.something = extcall Foo(msg.sender).foo()
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!
@internal
def foo():
    lib1.write_tuple()
    """
    main = """
import lib1
import lib2

initializes: lib1
initializes: lib2

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_function_call(make_input_bundle):
    # test missing uses through function call
    lib1 = """
counter: uint256

@internal
def update_counter(new_value: uint256):
    self.counter = new_value
    """
    lib2 = """
import lib1

# forgot `uses: lib1`!

counter: uint256

@internal
def foo():
    lib1.update_counter(lib1.counter + 1)
    """
    main = """
import lib1
import lib2

initializes: lib2
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_nested_attribute(make_input_bundle):
    # test missing uses through nested attribute access
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

counter: uint256

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib1

# did not `use` or `initialize` lib2!

@external
def foo(new_value: uint256):
    # cannot access lib1 state through lib2
    lib2.lib1.counter = new_value
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib2` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib2` or `initializes: lib2` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_subscript(make_input_bundle):
    # test missing uses through nested subscript/attribute access
    lib1 = """
struct Foo:
    array: uint256[5]

foos: Foo[5]
    """
    lib2 = """
import lib1

counter: uint256

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib1

# did not `use` or `initialize` lib2!

@external
def foo(new_value: uint256):
    # cannot access lib1 state through lib2
    lib2.lib1.foos[0].array[1] = new_value
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib2` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib2` or `initializes: lib2` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_missing_uses_nested_attribute_function_call(make_input_bundle):
    # test missing uses through nested attribute access
    lib1 = """
counter: uint256

@internal
def update_counter(new_value: uint256):
    self.counter = new_value
    """
    lib2 = """
import lib1

counter: uint256

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib1

# did not `use` or `initialize` lib2!

@external
def foo(new_value: uint256):
    # cannot access lib1 state through lib2
    lib2.lib1.update_counter(new_value)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib2` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib2` or `initializes: lib2` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_uses_skip_import(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib2

@external
def foo(new_value: uint256):
    # cannot access lib1 state through lib2, lib2 does not `use` lib1.
    lib2.lib1.counter = new_value
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value._hint == expected_hint


def test_uses_skip_import2(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

initializes: lib1

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib2

@external
def foo(new_value: uint256):
    # *can* access lib1 state through lib2, because lib2 initializes lib1
    lib2.lib1.counter = new_value
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_invalid_uses(make_input_bundle, chdir_tmp_path):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

uses: lib1  # not necessary!

counter: uint256

@internal
def foo():
    pass
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(BorrowException) as e:
        compile_code(main, input_bundle=input_bundle)
    expected = "`lib1` is declared as used, but its state is not"
    expected += " actually used in lib2.vy!"
    assert e.value._message == expected
    assert e.value._hint == "delete `uses: lib1`"


def test_invalid_uses2(make_input_bundle, chdir_tmp_path):
    # test a more complicated invalid uses
    lib1 = """
counter: uint256

@internal
def foo(addr: address):
    # sends value -- modifies ethereum state
    to_send_value: uint256 = 100
    raw_call(addr, b"someFunction()", value=to_send_value)
    """
    lib2 = """
import lib1

uses: lib1  # not necessary!

counter: uint256

@internal
def foo():
    lib1.foo(msg.sender)
    """
    main = """
import lib1
import lib2

initializes: lib2[lib1 := lib1]
initializes: lib1

@external
def foo():
    lib2.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    with pytest.raises(BorrowException) as e:
        compile_code(main, input_bundle=input_bundle)
    expected = "`lib1` is declared as used, but its state is not "
    expected += "actually used in lib2.vy!"
    assert e.value._message == expected
    assert e.value._hint == "delete `uses: lib1`"


def test_initializes_uses_conflict(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1

initializes: lib1
uses: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "ownership already set to `initializes`"


def test_uses_initializes_conflict(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1

uses: lib1
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "ownership already set to `uses`"


def test_uses_twice(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1

uses: lib1

random_variable: constant(uint256) = 3

uses: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "ownership already set to `uses`"


def test_initializes_twice(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1

initializes: lib1

random_variable: constant(uint256) = 3

initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "ownership already set to `initializes`"


def test_no_initialize_unused_module(make_input_bundle):
    lib1 = """
counter: uint256

@internal
def set_counter(new_value: uint256):
    self.counter = new_value

@internal
@pure
def add(x: uint256, y: uint256) -> uint256:
    return x + y
    """
    main = """
import lib1

# not needed: `initializes: lib1`

@external
def do_add(x: uint256, y: uint256) -> uint256:
    return lib1.add(x, y)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    assert compile_code(main, input_bundle=input_bundle) is not None


def test_no_initialize_unused_module2(make_input_bundle):
    # slightly more complicated
    lib1 = """
counter: uint256

@internal
def set_counter(new_value: uint256):
    self.counter = new_value

@internal
@pure
def add(x: uint256, y: uint256) -> uint256:
    return x + y
    """
    lib2 = """
import lib1

@internal
@pure
def addmul(x: uint256, y: uint256, z: uint256) -> uint256:
    return lib1.add(x, y) * z
    """
    main = """
import lib1
import lib2

@external
def do_addmul(x: uint256, y: uint256) -> uint256:
    return lib2.addmul(x, y, 5)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    assert compile_code(main, input_bundle=input_bundle) is not None


def test_init_uninitialized_function(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    main = """
import lib1

# missing `initializes: lib1`!

@deploy
def __init__():
    lib1.__init__()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "tried to initialize `lib1`, but it is not in initializer list!"
    assert e.value._hint == "add `initializes: lib1` as a top-level statement to your contract"


def test_init_uninitialized_function2(make_input_bundle):
    # test that we can't call module.__init__() even when we call `uses`
    lib1 = """
counter: uint256

@deploy
def __init__():
    pass
    """
    main = """
import lib1

uses: lib1
# missing `initializes: lib1`!

@deploy
def __init__():
    lib1.__init__()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "tried to initialize `lib1`, but it is not in initializer list!"
    assert e.value._hint == "add `initializes: lib1` as a top-level statement to your contract"


def test_noinit_initialized_function(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    self.counter = 5
    """
    main = """
import lib1

initializes: lib1

@deploy
def __init__():
    pass  # missing `lib1.__init__()`!
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "not initialized!"
    assert e.value._hint == "add `lib1.__init__()` to your `__init__()` function"


def test_noinit_initialized_function2(make_input_bundle):
    lib1 = """
counter: uint256

@deploy
def __init__():
    self.counter = 5
    """
    main = """
import lib1

initializes: lib1

# missing `lib1.__init__()`!
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "not initialized!"
    assert e.value._hint == "add `lib1.__init__()` to your `__init__()` function"


def test_ownership_decl_errors_not_swallowed(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1
# forgot to import lib2

uses: (lib1, lib2)  # should get UndeclaredDefinition
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(UndeclaredDefinition) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "'lib2' has not been declared."


def test_partial_compilation(make_input_bundle):
    lib1 = """
counter: uint256
    """
    main = """
import lib1

uses: lib1

@internal
def use_lib1():
    lib1.counter += 1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    assert (
        compile_code(main, input_bundle=input_bundle, output_formats=["annotated_ast_dict"])
        is not None
    )


def test_hint_for_missing_initializer_in_list(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib3 = """
counter: uint256
        """
    lib2 = """
import lib1
import lib3

uses: lib1
uses: lib3

counter: uint256

@internal
def foo():
    lib1.counter += 1
    lib3.counter += 1
    """
    main = """
import lib1
import lib2
import lib3

initializes: lib2[lib1:=lib1]
initializes: lib1
initializes: lib3
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "lib3.vy": lib3})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "`lib2` uses `lib3`, but it is not initialized with `lib3`"
    assert e.value._hint == "add `lib3 := lib3` to its initializer list"


def test_hint_for_missing_initializer_when_no_import(make_input_bundle):
    lib1 = """
counter: uint256
    """
    lib2 = """
import lib1

uses: lib1

counter: uint256

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib2

initializes: lib2
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    with pytest.raises(InitializerException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "`lib2` uses `lib1`, but it is not initialized with `lib1`"
    assert e.value._hint == "try importing lib1 first"


@pytest.fixture
def nonreentrant_library_bundle(make_input_bundle):
    # test simple case
    lib1 = """
# lib1.vy
@internal
@nonreentrant
def bar():
    pass

# lib1.vy
@external
@nonreentrant
def ext_bar():
    pass
    """
    # test case with recursion
    lib2 = """
@internal
def bar():
    self.baz()

@external
def ext_bar():
    self.baz()

@nonreentrant
@internal
def baz():
    return
    """
    # test case with nested recursion
    lib3 = """
import lib1
uses: lib1

@internal
def bar():
    lib1.bar()

@external
def ext_bar():
    lib1.bar()
    """

    return make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "lib3.vy": lib3})


@pytest.mark.parametrize("lib", ("lib1", "lib2", "lib3"))
def test_nonreentrant_exports(nonreentrant_library_bundle, lib):
    main = f"""
import {lib}

exports: {lib}.ext_bar  # line 4

@external
def foo():
    pass
    """
    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=nonreentrant_library_bundle)
    assert e.value._message == f"Cannot access `{lib}` state!" + NONREENTRANT_NOTE
    hint = f"add `uses: {lib}` or `initializes: {lib}` as a top-level statement to your contract"
    assert e.value._hint == hint
    assert e.value.annotations[0].lineno == 4


@pytest.mark.parametrize("lib", ("lib1", "lib2", "lib3"))
def test_internal_nonreentrant_import(nonreentrant_library_bundle, lib):
    main = f"""
import {lib}

@external
def foo():
    {lib}.bar()  # line 6
    """
    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=nonreentrant_library_bundle)
    assert e.value._message == f"Cannot access `{lib}` state!" + NONREENTRANT_NOTE

    hint = f"add `uses: {lib}` or `initializes: {lib}` as a top-level statement to your contract"
    assert e.value._hint == hint
    assert e.value.annotations[0].lineno == 6



def test_uint256_addmod(tx_failed, get_contract):
    uint256_code = """
@external
def _uint256_addmod(x: uint256, y: uint256, z: uint256) -> uint256:
    return uint256_addmod(x, y, z)
    """

    c = get_contract(uint256_code)

    assert c._uint256_addmod(1, 2, 2) == 1
    assert c._uint256_addmod(32, 2, 32) == 2
    assert c._uint256_addmod((2**256) - 1, 0, 2) == 1
    assert c._uint256_addmod(2**255, 2**255, 6) == 4
    with tx_failed():
        c._uint256_addmod(1, 2, 0)


def test_uint256_addmod_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: uint256) -> uint256: payable

@external
def foo(f: Foo) -> uint256:
    return uint256_addmod(32, 2, extcall f.foo(32))
    """

    c1 = side_effects_contract("uint256")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == 2
    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_uint256_addmod_internal_call(get_contract):
    code = """
@external
def foo() -> uint256:
    return uint256_addmod(self.a(), self.b(), self.c())

@internal
def a() -> uint256:
    return 32

@internal
def b() -> uint256:
    return 2

@internal
def c() -> uint256:
    return 32
    """

    c = get_contract(code)

    assert c.foo() == 2


def test_uint256_addmod_evaluation_order(get_contract):
    code = """
a: uint256

@external
def foo1() -> uint256:
    self.a = 0
    return uint256_addmod(self.a, 1, self.bar())

@external
def foo2() -> uint256:
    self.a = 0
    return uint256_addmod(self.a, self.bar(), 3)

@external
def foo3() -> uint256:
    self.a = 0
    return uint256_addmod(1, self.a, self.bar())

@internal
def bar() -> uint256:
    self.a = 1
    return 2
    """

    c = get_contract(code)

    assert c.foo1() == 1
    assert c.foo2() == 2
    assert c.foo3() == 1



from vyper.codegen.abi_encoder import abi_encoding_matches_vyper
from vyper.codegen.context import Context, VariableRecord
from vyper.codegen.core import get_element_ptr, make_setter, needs_clamp
from vyper.codegen.expr import Expr
from vyper.codegen.function_definitions.common import (
    EntryPointInfo,
    ExternalFuncIR,
    get_nonreentrant_lock,
    initialize_context,
    tag_frame_info,
)
from vyper.codegen.ir_node import Encoding, IRnode
from vyper.codegen.stmt import parse_body
from vyper.evm.address_space import CALLDATA, DATA
from vyper.semantics.types import TupleT
from vyper.semantics.types.function import ContractFunctionT
from vyper.utils import calc_mem_gas


# register function args with the local calling context.
# also allocate the ones that live in memory (i.e. kwargs)
def _register_function_args(func_t: ContractFunctionT, context: Context) -> list[IRnode]:
    ret = []
    # the type of the calldata
    base_args_t = TupleT(tuple(arg.typ for arg in func_t.positional_args))

    # tuple with the abi_encoded args
    if func_t.is_constructor:
        base_args_ofst = IRnode(0, location=DATA, typ=base_args_t, encoding=Encoding.ABI)
    else:
        base_args_ofst = IRnode(4, location=CALLDATA, typ=base_args_t, encoding=Encoding.ABI)

    for i, arg in enumerate(func_t.positional_args):
        arg_ir = get_element_ptr(base_args_ofst, i)

        if needs_clamp(arg.typ, Encoding.ABI):
            # allocate a memory slot for it and copy
            dst = context.new_variable(arg.name, arg.typ, is_mutable=False)

            copy_arg = make_setter(dst, arg_ir)
            copy_arg.ast_source = arg.ast_source
            ret.append(copy_arg)
        else:
            assert abi_encoding_matches_vyper(arg.typ)
            # leave it in place
            context.vars[arg.name] = VariableRecord(
                name=arg.name,
                pos=arg_ir,
                typ=arg.typ,
                mutable=False,
                location=arg_ir.location,
                encoding=Encoding.ABI,
            )

    return ret


def _generate_kwarg_handlers(
    func_t: ContractFunctionT, context: Context
) -> dict[str, EntryPointInfo]:
    # generate kwarg handlers.
    # since they might come in thru calldata or be default,
    # allocate them in memory and then fill it in based on calldata or default,
    # depending on the ContractFunctionT
    # a kwarg handler looks like
    # (if (eq _method_id <method_id>)
    #    copy calldata args to memory
    #    write default args to memory
    #    goto external_function_common_ir

    def handler_for(calldata_kwargs, default_kwargs):
        calldata_args = func_t.positional_args + calldata_kwargs
        # create a fake type so that get_element_ptr works
        calldata_args_t = TupleT(list(arg.typ for arg in calldata_args))

        abi_sig = func_t.abi_signature_for_kwargs(calldata_kwargs)

        calldata_kwargs_ofst = IRnode(
            4, location=CALLDATA, typ=calldata_args_t, encoding=Encoding.ABI
        )

        # a sequence of statements to strictify kwargs into memory
        ret = ["seq"]

        # ensure calldata is at least of minimum length
        args_abi_t = calldata_args_t.abi_type
        calldata_min_size = args_abi_t.min_size() + 4

        # TODO optimize make_setter by using
        # TupleT(list(arg.typ for arg in calldata_kwargs + default_kwargs))
        # (must ensure memory area is contiguous)

        for i, arg_meta in enumerate(calldata_kwargs):
            k = func_t.n_positional_args + i

            lhs = context.lookup_var(arg_meta.name).as_ir_node()

            rhs = get_element_ptr(calldata_kwargs_ofst, k, array_bounds_check=False)

            copy_arg = make_setter(lhs, rhs)
            copy_arg.ast_source = arg_meta.ast_source
            ret.append(copy_arg)

        for x in default_kwargs:
            lhs = context.lookup_var(x.name).as_ir_node()
            lhs.ast_source = x.ast_source
            kw_ast_val = func_t.default_values[x.name]  # e.g. `3` in x: int = 3
            rhs = Expr(kw_ast_val, context).ir_node

            copy_arg = make_setter(lhs, rhs)
            copy_arg.ast_source = x.ast_source
            ret.append(copy_arg)

        ret.append(["goto", func_t._ir_info.external_function_base_entry_label])

        # return something we can turn into ExternalFuncIR
        return abi_sig, calldata_min_size, ret

    ret = {}

    keyword_args = func_t.keyword_args

    # allocate variable slots in memory
    for arg in keyword_args:
        context.new_variable(arg.name, arg.typ, is_mutable=False)

    for i, _ in enumerate(keyword_args):
        calldata_kwargs = keyword_args[:i]
        default_kwargs = keyword_args[i:]

        sig, calldata_min_size, ir_node = handler_for(calldata_kwargs, default_kwargs)
        assert sig not in ret
        ret[sig] = EntryPointInfo(func_t, calldata_min_size, ir_node)

    sig, calldata_min_size, ir_node = handler_for(keyword_args, [])

    assert sig not in ret
    ret[sig] = EntryPointInfo(func_t, calldata_min_size, ir_node)

    return ret


def _adjust_gas_estimate(func_t, common_ir):
    # adjust gas estimate to include cost of mem expansion
    # frame_size of external function includes all private functions called
    # (note: internal functions do not need to adjust gas estimate since
    frame_info = func_t._ir_info.frame_info

    mem_expansion_cost = calc_mem_gas(frame_info.mem_used)
    common_ir.add_gas_estimate += mem_expansion_cost
    func_t._ir_info.gas_estimate = common_ir.gas


def generate_ir_for_external_function(code, compilation_target):
    # TODO type hints:
    # def generate_ir_for_external_function(
    #    code: vy_ast.FunctionDef,
    #    compilation_target: ModuleT,
    # ) -> IRnode:
    """
    Return the IR for an external function. Returns IR for the body
    of the function, handle kwargs and exit the function. Also returns
    metadata required for `module.py` to construct the selector table.
    """
    func_t = code._metadata["func_type"]
    assert func_t.is_external or func_t.is_constructor  # sanity check

    context = initialize_context(func_t, compilation_target, func_t.is_constructor)
    nonreentrant_pre, nonreentrant_post = get_nonreentrant_lock(func_t)

    # generate handlers for base args and register the variable records
    handle_base_args = _register_function_args(func_t, context)

    # generate handlers for kwargs and register the variable records
    entry_points = _generate_kwarg_handlers(func_t, context)

    body = ["seq"]
    # once optional args have been handled,
    # generate the main body of the function
    body += handle_base_args

    body += nonreentrant_pre

    body += [parse_body(code.body, context, ensure_terminated=True)]

    # wrap the body in labeled block
    body = ["label", func_t._ir_info.external_function_base_entry_label, ["var_list"], body]

    exit_sequence = ["seq"] + nonreentrant_post
    if func_t.is_constructor:
        pass  # init func has special exit sequence generated by module.py
    elif context.return_type is None:
        exit_sequence += [["stop"]]
    else:
        exit_sequence += [["return", "ret_ofst", "ret_len"]]

    exit_sequence_args = ["var_list"]
    if context.return_type is not None:
        exit_sequence_args += ["ret_ofst", "ret_len"]
    # wrap the exit in a labeled block
    exit_ = ["label", func_t._ir_info.exit_sequence_label, exit_sequence_args, exit_sequence]

    # the ir which comprises the main body of the function,
    # besides any kwarg handling
    func_common_ir = IRnode.from_list(["seq", body, exit_], ast_source=code)

    tag_frame_info(func_t, context)

    _adjust_gas_estimate(func_t, func_common_ir)

    return ExternalFuncIR(entry_points, func_common_ir)



from vyper.codegen.core import _freshname, eval_once_check, make_setter
from vyper.codegen.ir_node import IRnode
from vyper.evm.address_space import MEMORY
from vyper.exceptions import StateAccessViolation
from vyper.semantics.types.subscriptable import TupleT


def _align_kwargs(func_t, args_ir):
    """
    Using a list of args, find the kwargs which need to be filled in by
    the compiler
    """

    # sanity check
    assert func_t.n_positional_args <= len(args_ir) <= func_t.n_total_args

    num_provided_kwargs = len(args_ir) - func_t.n_positional_args

    unprovided_kwargs = func_t.keyword_args[num_provided_kwargs:]
    return [i.default_value for i in unprovided_kwargs]


def ir_for_self_call(stmt_expr, context):
    from vyper.codegen.expr import Expr  # TODO rethink this circular import

    # ** Internal Call **
    # Steps:
    # - copy arguments into the soon-to-be callee
    # - allocate return buffer
    # - push jumpdest (callback ptr) and return buffer location
    # - jump to label
    # - (private function will fill return buffer and jump back)
    method_name = stmt_expr.func.attr
    func_t = stmt_expr.func._metadata["type"]

    pos_args_ir = [Expr(x, context).ir_node for x in stmt_expr.args]

    default_vals = _align_kwargs(func_t, pos_args_ir)
    default_vals_ir = [Expr(x, context).ir_node for x in default_vals]

    args_ir = pos_args_ir + default_vals_ir
    assert len(args_ir) == len(func_t.arguments)

    args_tuple_t = TupleT([x.typ for x in args_ir])
    args_as_tuple = IRnode.from_list(["multi"] + [x for x in args_ir], typ=args_tuple_t)

    # CMC 2023-05-17 this seems like it is already caught in typechecker
    if context.is_constant() and func_t.is_mutable:
        raise StateAccessViolation(
            f"May not call state modifying function "
            f"'{method_name}' within {context.pp_constancy()}.",
            stmt_expr,
        )

    # note: internal_function_label asserts `func_t.is_internal`.
    _label = func_t._ir_info.internal_function_label(context.is_ctor_context)
    return_label = _freshname(f"{_label}_call")

    # allocate space for the return buffer
    # TODO allocate in stmt and/or expr.py
    if func_t.return_type is not None:
        return_buffer = IRnode.from_list(
            context.new_internal_variable(func_t.return_type),
            annotation=f"{return_label}_return_buf",
        )
    else:
        return_buffer = None

    # note: dst_tuple_t != args_tuple_t
    dst_tuple_t = TupleT(tuple(func_t.argument_types))
    args_dst = IRnode(func_t._ir_info.frame_info.frame_start, typ=dst_tuple_t, location=MEMORY)

    # if one of the arguments is a self call, the argument
    # buffer could get borked. to prevent against that,
    # write args to a temporary buffer until all the arguments
    # are fully evaluated.
    if args_as_tuple.contains_self_call:
        copy_args = ["seq"]
        # TODO deallocate me
        tmp_args_buf = context.new_internal_variable(dst_tuple_t)
        copy_args.append(
            # --> args evaluate here <--
            make_setter(tmp_args_buf, args_as_tuple)
        )

        copy_args.append(make_setter(args_dst, tmp_args_buf))

    else:
        copy_args = make_setter(args_dst, args_as_tuple)

    goto_op = ["goto", func_t._ir_info.internal_function_label(context.is_ctor_context)]
    # pass return buffer to subroutine
    if return_buffer is not None:
        goto_op += [return_buffer]
    # pass return label to subroutine
    goto_op.append(["symbol", return_label])

    call_sequence = ["seq"]
    call_sequence.append(eval_once_check(_freshname(stmt_expr.node_source_code)))
    call_sequence.extend([copy_args, goto_op, ["label", return_label, ["var_list"], "pass"]])
    if return_buffer is not None:
        # push return buffer location to stack
        call_sequence += [return_buffer]

    o = IRnode.from_list(
        call_sequence,
        typ=func_t.return_type,
        location=MEMORY,
        annotation=stmt_expr.get("node_source_code"),
        add_gas_estimate=func_t._ir_info.gas_estimate,
    )
    o.is_self_call = True
    return o



import pytest

from vyper import compile_code
from vyper.exceptions import (
    ArgumentException,
    InvalidLiteral,
    OverflowException,
    StructureException,
    TypeMismatch,
    UndeclaredDefinition,
)

# CMC 2023-12-31 these tests could probably go in builtins/folding/

fail_list = [
    (
        """
@external
def foo():
    x: uint256 = as_wei_value(5, szabo)
    """,
        UndeclaredDefinition,
    ),
    (
        """
@external
def foo():
    x: uint256 = as_wei_value(5, "szaboo")
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo() -> int128:
    x: int128 = 45
    return x.balance
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    x: int128 = as_wei_value(0xf5, "szabo")
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> uint256:
    return as_wei_value(
        115792089237316195423570985008687907853269984665640564039457584007913129639937,
        'milliether'
    )
    """,
        OverflowException,
    ),
    (
        """
@external
def foo():
    x: uint256 = as_wei_value(-1, "szabo")
    """,
        InvalidLiteral,
    ),
    (
        """
FOO: constant(uint256) = as_wei_value(5, szabo)
    """,
        UndeclaredDefinition,
    ),
    (
        """
FOO: constant(uint256) = as_wei_value(5, "szaboo")
    """,
        ArgumentException,
    ),
    (
        """
FOO: constant(uint256) = as_wei_value(-1, "szabo")
    """,
        InvalidLiteral,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_as_wei_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    x: uint256 = as_wei_value(5, "finney") + as_wei_value(2, "babbage") + as_wei_value(8, "shannon")  # noqa: E501
    """,
    """
@external
def foo():
    z: int128 = 2 + 3
    x: uint256 = as_wei_value(2 + 3, "finney")
    """,
    """
@external
def foo():
    x: uint256 = as_wei_value(5.182, "babbage")
    """,
    """
@external
def foo() -> uint256:
    x: address = 0x1234567890123456789012345678901234567890
    return x.balance
    """,
    """
y: constant(String[5]) = "szabo"
x: constant(uint256) = as_wei_value(5, y)

@external
def foo():
    a: uint256 = x
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_as_wei_success(good_code, get_contract):
    assert get_contract(good_code) is not None



import pytest

from tests.utils import decimal_to_int
from vyper.exceptions import InvalidOperation


def test_unary_sub_uint256_fail(assert_compile_failed, get_contract):
    code = """@external
def negate(a: uint256) -> uint256:
    return -(a)
    """
    assert_compile_failed(lambda: get_contract(code), exception=InvalidOperation)


def test_unary_sub_int128_fail(get_contract, tx_failed):
    code = """@external
def negate(a: int128) -> int128:
    return -(a)
    """
    c = get_contract(code)
    # This test should revert on overflow condition
    with tx_failed():
        c.negate(-(2**127))


@pytest.mark.parametrize("val", [-(2**127) + 1, 0, 2**127 - 1])
def test_unary_sub_int128_pass(get_contract, val):
    code = """@external
def negate(a: int128) -> int128:
    return -(a)
    """
    c = get_contract(code)
    assert c.negate(val) == -val


min_decimal = -(2**127) + 1
max_decimal = 2**127 - 1


@pytest.mark.parametrize("val", [min_decimal, 0, max_decimal])
def test_unary_sub_decimal_pass(get_contract, val):
    code = """@external
def negate(a: decimal) -> decimal:
    return -(a)
    """
    c = get_contract(code)
    assert c.negate(val) == -val


def test_negation_decimal(get_contract):
    code = """
a: constant(decimal) = 18707220957835557353007165858768422651595.9365500927
b: constant(decimal) = -18707220957835557353007165858768422651595.9365500927

@external
def foo() -> decimal:
    return -a

@external
def bar() -> decimal:
    return -b
    """

    c = get_contract(code)
    assert c.foo() == decimal_to_int("-18707220957835557353007165858768422651595.9365500927")
    assert c.bar() == decimal_to_int("18707220957835557353007165858768422651595.9365500927")


def test_negation_int128(get_contract):
    code = """
a: constant(int128) = min_value(int128)

@external
def bar() -> int128:
    return -(a+1)
    """
    c = get_contract(code)
    assert c.bar() == 2**127 - 1



import copy

from vyper import ast as vy_ast
from vyper.exceptions import CompilerPanic
from vyper.semantics.types.function import ContractFunctionT


# CMC 2024-02-20 NOTE: it may be beneficial to remove this function; it
# causes correctness and performance problems because of copying and mutating
# the AST. we could handle getter generation during code generation.
def generate_public_variable_getters(vyper_module: vy_ast.Module) -> None:
    """
    Create getter functions for public variables.

    Arguments
    ---------
    vyper_module : Module
        Top-level Vyper AST node.
    """

    for node in vyper_module.get_children(vy_ast.VariableDecl, {"is_public": True}):
        funcname = node.target.id
        input_types, return_type = node._metadata["type"].getter_signature

        input_nodes = []

        # use the annotation node to build the input args and return type
        annotation = copy.copy(node.annotation)

        return_expr: vy_ast.ExprNode
        # constants just return a value
        if node.is_constant:
            return_expr = node.value
        elif node.is_immutable:
            return_expr = vy_ast.Name(id=funcname)  # type: ignore
        else:
            # the base return statement is an `Attribute` node, e.g.
            # `self.<var_name>`. for each input type we wrap it in a
            # `Subscript` to access a specific member
            return_expr = vy_ast.Attribute(value=vy_ast.Name(id="self"), attr=funcname)

        for i, type_ in enumerate(input_types):
            if not isinstance(annotation, vy_ast.Subscript):
                # if we get here something has failed in type checking
                raise CompilerPanic("Mismatch between node and input type while building getter")
            if annotation.value.get("id") == "HashMap":  # type: ignore
                # for a HashMap, split the key/value types and use the key
                # type as the next arg
                arg, annotation = annotation.slice.elements  # type: ignore
            elif annotation.value.get("id") == "DynArray":
                arg = vy_ast.Name(id=type_._id)
                annotation = annotation.slice.elements[0]  # type: ignore
            else:
                # for other types, build an input arg node from the expected
                # type and remove the outer `Subscript` from the annotation
                arg = vy_ast.Name(id=type_._id)
                annotation = annotation.value
            input_nodes.append(vy_ast.arg(arg=f"arg{i}", annotation=arg))

            # wrap the return expression in a `Subscript`
            return_expr = vy_ast.Subscript(value=return_expr, slice=vy_ast.Name(id=f"arg{i}"))

        # after iterating the input types, the remaining annotation node is our return type
        return_annotation = copy.copy(annotation)

        # join everything together as a new `FunctionDef` node
        expanded = vy_ast.FunctionDef(
            name=funcname,
            args=vy_ast.arguments(args=input_nodes, defaults=[]),
            body=[vy_ast.Return(value=return_expr)],
            decorator_list=[vy_ast.Name(id="external"), vy_ast.Name(id="view")],
            returns=return_annotation,
        )

        # set some pointers for error messages
        expanded._original_node = node
        expanded.set_parent(node.parent)

        func_type = ContractFunctionT.from_FunctionDef(expanded)
        expanded._metadata["func_type"] = func_type

        # update pointers
        node._expanded_getter = expanded



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    a: uint256 = pow_mod256(-1, -1)
    """,
        TypeMismatch,
    )
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_powmod_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
FOO: constant(uint256) = 3
BAR: constant(uint256) = 5
BAZ: constant(uint256) = pow_mod256(FOO, BAR)

@external
def foo():
    a: uint256 = BAZ
    """
]


@pytest.mark.parametrize("code", valid_list)
def test_powmod_pass(code):
    assert compile_code(code) is not None



import enum
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Union

from vyper import ast as vy_ast
from vyper.compiler.input_bundle import CompilerInput
from vyper.exceptions import CompilerPanic, StructureException
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types.base import VyperType
from vyper.semantics.types.primitives import SelfT
from vyper.utils import OrderedSet, StringEnum

if TYPE_CHECKING:
    from vyper.semantics.types.function import ContractFunctionT
    from vyper.semantics.types.module import InterfaceT, ModuleT


class FunctionVisibility(StringEnum):
    EXTERNAL = enum.auto()
    INTERNAL = enum.auto()
    DEPLOY = enum.auto()


class StateMutability(StringEnum):
    PURE = enum.auto()
    VIEW = enum.auto()
    NONPAYABLE = enum.auto()
    PAYABLE = enum.auto()

    @classmethod
    def from_abi(cls, abi_dict: Dict) -> "StateMutability":
        """
        Extract stateMutability from an entry in a contract's ABI
        """
        if "stateMutability" in abi_dict:
            return cls(abi_dict["stateMutability"])
        elif abi_dict.get("payable"):
            return StateMutability.PAYABLE
        elif "constant" in abi_dict and abi_dict["constant"]:
            return StateMutability.VIEW
        else:  # Assume nonpayable if neither field is there, or constant/payable not set
            return StateMutability.NONPAYABLE
        # NOTE: The state mutability nonpayable is reflected in Solidity by not
        #       specifying a state mutability modifier at all. Do the same here.


# classify the constancy of an expression
# CMC 2023-12-31 note that we now have three ways of classifying mutability in
# the codebase: StateMutability (for functions), Modifiability (for expressions
# and variables) and Constancy (in codegen). context.Constancy can/should
# probably be refactored away though as those kinds of checks should be done
# during analysis.
class Modifiability(StringEnum):
    # compile-time / always constant
    CONSTANT = enum.auto()

    # things that are constant within the current message call, including
    # block.*, msg.*, tx.* and immutables
    RUNTIME_CONSTANT = enum.auto()

    # could potentially add more fine-grained here as needed, like
    # CONSTANT_AFTER_DEPLOY, TX_CONSTANT, BLOCK_CONSTANT, etc.

    # is writeable/can result in arbitrary state or memory changes
    MODIFIABLE = enum.auto()

    @classmethod
    def from_state_mutability(cls, mutability: StateMutability):
        if mutability == StateMutability.PURE:
            return cls.CONSTANT
        if mutability == StateMutability.VIEW:
            return cls.RUNTIME_CONSTANT
        # sanity check in case more StateMutability levels are added in the future
        assert mutability in (StateMutability.PAYABLE, StateMutability.NONPAYABLE)
        return cls.MODIFIABLE


@dataclass
class VarOffset:
    position: int


class ModuleOwnership(StringEnum):
    NO_OWNERSHIP = enum.auto()  # readable
    USES = enum.auto()  # writeable
    INITIALIZES = enum.auto()  # initializes


# base class for things that are the "result" of analysis
class AnalysisResult:
    pass


@dataclass
class ModuleInfo(AnalysisResult):
    module_t: "ModuleT"
    alias: str
    ownership: ModuleOwnership = ModuleOwnership.NO_OWNERSHIP
    ownership_decl: Optional[vy_ast.VyperNode] = None

    @property
    def module_node(self):
        return self.module_t._module

    # duck type, conform to interface of VarInfo and ExprInfo
    @property
    def typ(self):
        return self.module_t

    def set_ownership(self, module_ownership: ModuleOwnership, node: Optional[vy_ast.VyperNode]):
        if self.ownership != ModuleOwnership.NO_OWNERSHIP:
            raise StructureException(
                f"ownership already set to `{self.ownership}`", node, self.ownership_decl
            )
        self.ownership = module_ownership

    def __hash__(self):
        return hash(id(self.module_t))


@dataclass(frozen=True)
class ImportInfo(AnalysisResult):
    typ: Union[ModuleInfo, "InterfaceT"]
    alias: str  # the name in the namespace
    qualified_module_name: str  # for error messages
    compiler_input: CompilerInput  # to recover file info for ast export
    node: vy_ast.VyperNode

    def to_dict(self):
        ret = {"alias": self.alias, "qualified_module_name": self.qualified_module_name}

        ret["source_id"] = self.compiler_input.source_id
        ret["path"] = str(self.compiler_input.path)
        ret["resolved_path"] = str(self.compiler_input.resolved_path)
        ret["file_sha256sum"] = self.compiler_input.sha256sum

        return ret


# analysis result of InitializesDecl
@dataclass
class InitializesInfo(AnalysisResult):
    module_info: ModuleInfo
    dependencies: list[ModuleInfo]
    node: Optional[vy_ast.VyperNode] = None


# analysis result of UsesDecl
@dataclass
class UsesInfo(AnalysisResult):
    used_modules: list[ModuleInfo]
    node: Optional[vy_ast.VyperNode] = None


# analysis result of ExportsDecl
@dataclass
class ExportsInfo(AnalysisResult):
    functions: list["ContractFunctionT"]
    used_modules: OrderedSet[ModuleInfo]


@dataclass
class VarInfo:
    """
    VarInfo are objects that represent the type of a variable,
    plus associated metadata like location and modifiability attributes

    Object Attributes
    -----------------
    location: DataLocation of this variable
    modifiability: Modifiability of this variable
    """

    typ: VyperType
    location: DataLocation = DataLocation.UNSET
    modifiability: Modifiability = Modifiability.MODIFIABLE
    is_public: bool = False
    decl_node: Optional[vy_ast.VariableDecl] = None

    def __hash__(self):
        return hash(id(self))

    def __post_init__(self):
        self.position = None
        self._modification_count = 0

    @property
    def getter_ast(self) -> Optional[vy_ast.VyperNode]:
        assert self.decl_node is not None  # help mypy
        ret = self.decl_node._expanded_getter
        assert (ret is not None) == self.is_public, self
        return ret

    def set_position(self, position: VarOffset) -> None:
        if self.position is not None:
            raise CompilerPanic("Position was already assigned")
        assert isinstance(position, VarOffset)  # sanity check
        self.position = position

    def is_state_variable(self):
        non_state_locations = (DataLocation.UNSET, DataLocation.MEMORY, DataLocation.CALLDATA)
        # `self` gets a VarInfo, but it is not considered a state
        # variable (it is magic), so we ignore it here.
        return self.location not in non_state_locations and not isinstance(self.typ, SelfT)

    def get_size(self) -> int:
        return self.typ.get_size_in(self.location)

    @property
    def is_transient(self):
        return self.location == DataLocation.TRANSIENT

    @property
    def is_immutable(self):
        return self.location == DataLocation.CODE

    @property
    def is_constant(self):
        res = self.location == DataLocation.UNSET
        assert res == (self.modifiability == Modifiability.CONSTANT)
        return res


@dataclass(frozen=True)
class VarAccess:
    variable: VarInfo
    path: tuple[str | object, ...]

    # A sentinel indicating a subscript access
    SUBSCRIPT_ACCESS: ClassVar[Any] = object()

    @cached_property
    def attrs(self):
        ret = []
        for s in self.path:
            if s is self.SUBSCRIPT_ACCESS:
                break
            ret.append(s)
        return tuple(ret)

    def contains(self, other):
        # VarAccess("v", ("a")) `contains` VarAccess("v", ("a", "b", "c"))
        sub_attrs = other.attrs[: len(self.attrs)]
        return self.variable == other.variable and sub_attrs == self.attrs

    def to_dict(self):
        var = self.variable
        if var.decl_node is None:
            # happens for builtins or `self` accesses
            return None

        # map SUBSCRIPT_ACCESS to `"$subscript_access"` (which is an identifier
        # which can't be constructed by the user)
        path = ["$subscript_access" if s is self.SUBSCRIPT_ACCESS else s for s in self.path]
        varname = var.decl_node.target.id

        decl_node = var.decl_node.get_id_dict()
        ret = {"name": varname, "decl_node": decl_node, "access_path": path}
        return ret


@dataclass
class ExprInfo:
    """
    Class which represents the analysis associated with an expression
    """

    typ: VyperType
    var_info: Optional[VarInfo] = None
    module_info: Optional[ModuleInfo] = None
    location: DataLocation = DataLocation.UNSET
    modifiability: Modifiability = Modifiability.MODIFIABLE
    attr: Optional[str] = None

    def __post_init__(self):
        should_match = ("typ", "location", "modifiability")
        if self.var_info is not None:
            for attr in should_match:
                if getattr(self.var_info, attr) != getattr(self, attr):
                    raise CompilerPanic(f"Bad analysis: non-matching {attr}: {self}")

        self._writes: OrderedSet[VarAccess] = OrderedSet()
        self._reads: OrderedSet[VarAccess] = OrderedSet()

    @classmethod
    def from_varinfo(cls, var_info: VarInfo, **kwargs) -> "ExprInfo":
        return cls(
            var_info.typ,
            var_info=var_info,
            location=var_info.location,
            modifiability=var_info.modifiability,
            **kwargs,
        )

    @classmethod
    def from_moduleinfo(cls, module_info: ModuleInfo, **kwargs) -> "ExprInfo":
        modifiability = Modifiability.RUNTIME_CONSTANT
        if module_info.ownership >= ModuleOwnership.USES:
            modifiability = Modifiability.MODIFIABLE

        return cls(
            module_info.module_t, module_info=module_info, modifiability=modifiability, **kwargs
        )

    def copy_with_type(self, typ: VyperType, **kwargs) -> "ExprInfo":
        """
        Return a copy of the ExprInfo but with the type set to something else
        """
        to_copy = ("location", "modifiability")
        fields = {k: getattr(self, k) for k in to_copy}
        for t in to_copy:
            assert t not in kwargs
        return self.__class__(typ=typ, **fields, **kwargs)



import contextlib
import decimal
import os

from vyper import ast as vy_ast
from vyper.semantics.analysis.constant_folding import constant_fold
from vyper.utils import DECIMAL_EPSILON, round_towards_zero

ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"


@contextlib.contextmanager
def working_directory(directory):
    tmp = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(tmp)


def parse_and_fold(source_code):
    ast = vy_ast.parse_to_ast(source_code)
    constant_fold(ast)
    return ast


def decimal_to_int(*args):
    s = decimal.Decimal(*args)
    return round_towards_zero(s / DECIMAL_EPSILON)



import itertools

import pytest

from tests.utils import ZERO_ADDRESS, decimal_to_int
from vyper.compiler import compile_code
from vyper.exceptions import TypeMismatch
from vyper.utils import MemoryPositions


def search_for_sublist(ir, sublist):
    _list = ir.to_list() if hasattr(ir, "to_list") else ir
    if _list == sublist:
        return True
    return isinstance(_list, list) and any(search_for_sublist(i, sublist) for i in _list)


def test_builtin_constants(get_contract):
    code = """
@external
def test_zaddress(a: address) -> bool:
    return a == empty(address)


@external
def test_empty_bytes32(a: bytes32) -> bool:
    return a == empty(bytes32)


@external
def test_int128(a: int128) -> (bool, bool):
    return a == max_value(int128), a == min_value(int128)


@external
def test_decimal(a: decimal) -> (bool, bool):
    return a == max_value(decimal), a == min_value(decimal)


@external
def test_uint256(a: uint256) -> bool:
    return a == max_value(uint256)


@external
def test_arithmetic(a: int128) -> int128:
    return max_value(int128) - a
    """

    c = get_contract(code)

    assert c.test_empty_bytes32(b"\x00" * 32) is True
    assert c.test_empty_bytes32(b"\x0F" * 32) is False

    assert c.test_zaddress(ZERO_ADDRESS) is True
    assert c.test_zaddress("0x0000000000000000000000000000000000000012") is False

    assert c.test_int128(2**127 - 1) == (True, False)
    assert c.test_int128(-(2**127)) == (False, True)
    assert c.test_int128(0) == (False, False)

    assert c.test_decimal(
        decimal_to_int("18707220957835557353007165858768422651595.9365500927")
    ) == (True, False)
    assert c.test_decimal(
        decimal_to_int("-18707220957835557353007165858768422651595.9365500928")
    ) == (False, True)
    assert c.test_decimal(decimal_to_int("0.1")) == (False, False)

    assert c.test_uint256(2**256 - 1) is True

    assert c.test_arithmetic(5000) == 2**127 - 1 - 5000


def test_builtin_constants_assignment(get_contract):
    code = """
@external
def foo() -> int128:
    bar: int128 = max_value(int128)
    return bar

@external
def goo() -> int128:
    bar: int128 = min_value(int128)
    return bar

@external
def hoo() -> bytes32:
    bar: bytes32 = empty(bytes32)
    return bar

@external
def joo() -> address:
    bar: address = empty(address)
    return bar

@external
def koo() -> decimal:
    bar: decimal = max_value(decimal)
    return bar

@external
def loo() -> decimal:
    bar: decimal = min_value(decimal)
    return bar

@external
def zoo() -> uint256:
    bar: uint256 = max_value(uint256)
    return bar
    """

    c = get_contract(code)

    assert c.foo() == 2**127 - 1
    assert c.goo() == -(2**127)

    assert c.hoo() == b"\x00" * 32

    assert c.joo() == ZERO_ADDRESS

    assert c.koo() == (2**167 - 1)
    assert c.loo() == -(2**167)

    assert c.zoo() == 2**256 - 1


def test_custom_constants(get_contract):
    code = """
X_VALUE: constant(uint256) = 33

@external
def test() -> uint256:
    return X_VALUE

@external
def test_add(a: uint256) -> uint256:
    return X_VALUE + a
    """
    c = get_contract(code)

    assert c.test() == 33
    assert c.test_add(7) == 40


# Would be nice to put this somewhere accessible, like in vyper.types or something
integer_types = ["uint8", "int128", "int256", "uint256"]


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_custom_constants_fail(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant({storage_type}) = 1

@external
def foo() -> {return_type}:
    return MY_CONSTANT
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


def test_constant_address(get_contract):
    code = """
OWNER: constant(address) = 0x0000000000000000000000000000000000000012

@external
def get_owner() -> address:
    return OWNER

@external
def is_owner() -> bool:
    if msg.sender == OWNER:
        return True
    else:
        return False
    """
    c = get_contract(code)

    assert c.get_owner() == "0x0000000000000000000000000000000000000012"
    assert c.is_owner() is False


def test_constant_bytes(get_contract):
    test_str = b"Alabama, Arkansas. I do love my ma and pa"
    code = f"""
X: constant(Bytes[100]) = b"{test_str.decode()}"

@external
def test() -> Bytes[100]:
    y: Bytes[100] = X

    return y
    """

    c = get_contract(code)

    assert c.test() == test_str


def test_constant_folds(experimental_codegen):
    some_prime = 10013677
    code = f"""
SOME_CONSTANT: constant(uint256) = 11 + 1
SOME_PRIME: constant(uint256) = {some_prime}

@external
def test() -> uint256:
    # calculate some constant which is really unlikely to be randomly
    # in bytecode
    ret: uint256 = 2**SOME_CONSTANT * SOME_PRIME
    return ret
    """
    ir = compile_code(code, output_formats=["ir"])["ir"]
    memory = "$alloca_64_32" if experimental_codegen else MemoryPositions.RESERVED_MEMORY
    search = ["mstore", [memory], [2**12 * some_prime]]
    assert search_for_sublist(ir, search)


def test_constant_lists(get_contract):
    code = """
BYTE32_LIST: constant(bytes32[2]) = [
    0x0000000000000000000000000000000000000000000000000000000000001321,
    0x0000000000000000000000000000000000000000000000000000000000001123
]

SPECIAL: constant(int128[3]) = [33, 44, 55]

@external
def test() -> bytes32:
    a: bytes32[2] = BYTE32_LIST
    return a[1]

@view
@external
def contains(a: int128) -> bool:
    return a in SPECIAL
    """

    c = get_contract(code)

    assert c.test()[-2:] == b"\x11\x23"

    assert c.contains(55) is True
    assert c.contains(44) is True
    assert c.contains(33) is True
    assert c.contains(3) is False



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold
from vyper.compiler import compile_code
from vyper.exceptions import ArrayIndexException


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(
    idx=st.integers(min_value=0, max_value=9),
    array=st.lists(st.integers(), min_size=10, max_size=10),
)
def test_subscript(get_contract, array, idx):
    source = """
@external
def foo(array: int128[10], idx: uint256) -> int128:
    return array[idx]
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{array}[{idx}]")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(array, idx) == new_node.value


def test_negative_index():
    source = """
@external
def foo(array: int128[10]) -> int128:
    return array[0 - 1]
    """
    with pytest.raises(ArrayIndexException):
        compile_code(source)


def test_oob_index():
    source = """
@external
def foo(array: int128[10]) -> int128:
    return array[9 + 1]
    """
    with pytest.raises(ArrayIndexException):
        compile_code(source)



import ast as python_ast
from typing import Any, Optional, Sequence, Type, Union

from .natspec import parse_natspec as parse_natspec
from .parse import parse_to_ast as parse_to_ast
from .parse import parse_to_ast_with_settings as parse_to_ast_with_settings
from .utils import ast_to_dict as ast_to_dict

NODE_BASE_ATTRIBUTES: Any
NODE_SRC_ATTRIBUTES: Any
DICT_AST_SKIPLIST: Any

def get_node(
    ast_struct: Union[dict, python_ast.AST], parent: Optional[VyperNode] = ...
) -> VyperNode: ...

class VyperNode:
    full_source_code: str = ...
    node_source_code: str = ...
    lineno: int = ...
    col_offset: int = ...
    end_lineno: int = ...
    end_col_offset: int = ...
    _metadata: dict = ...
    _original_node: Optional[VyperNode] = ...
    def __init__(self, parent: Optional[VyperNode] = ..., **kwargs: Any) -> None: ...
    def __hash__(self) -> Any: ...
    def __eq__(self, other: Any) -> Any: ...
    @property
    def description(self): ...
    @property
    def is_literal_value(self): ...
    @property
    def has_folded_value(self): ...
    @property
    def parent(self): ...
    @classmethod
    def get_fields(cls: Any) -> set: ...
    def set_parent(self, parent: VyperNode) -> VyperNode: ...
    def get_folded_value(self) -> ExprNode: ...
    def reduced(self) -> ExprNode: ...
    def _set_folded_value(self, node: ExprNode) -> None: ...
    @classmethod
    def from_node(cls, node: VyperNode, **kwargs: Any) -> Any: ...
    def to_dict(self) -> dict: ...
    def get_children(
        self,
        node_type: Union[Type[VyperNode], Sequence[Type[VyperNode]], None] = ...,
        filters: Optional[dict] = ...,
        reverse: bool = ...,
    ) -> list: ...
    def get_descendants(
        self,
        node_type: Union[Type[VyperNode], Sequence[Type[VyperNode]], None] = ...,
        filters: Optional[dict] = ...,
        include_self: bool = ...,
        reverse: bool = ...,
    ) -> list: ...
    def get_ancestor(
        self, node_type: Union[Type[VyperNode], Sequence[Type[VyperNode]], None] = ...
    ) -> VyperNode: ...
    def get(self, field_str: str) -> Any: ...

class TopLevel(VyperNode):
    doc_string: Str = ...
    body: list = ...
    name: str = ...

class Module(TopLevel):
    path: str = ...
    resolved_path: str = ...
    def namespace(self) -> Any: ...  # context manager

class FunctionDef(TopLevel):
    args: arguments = ...
    decorator_list: list = ...
    returns: VyperNode = ...

class arguments(VyperNode):
    args: list = ...
    defaults: list = ...

class arg(VyperNode): ...
class Return(VyperNode): ...

class Log(VyperNode):
    value: Call = ...

class FlagDef(VyperNode):
    body: list = ...
    name: str = ...

class EventDef(VyperNode):
    body: list = ...
    name: str = ...

class InterfaceDef(VyperNode):
    body: list = ...
    name: str = ...

class StructDef(VyperNode):
    body: list = ...
    name: str = ...

class ExprNode(VyperNode):
    _expr_info: Any = ...

class Constant(ExprNode):
    value: Any = ...

class Num(Constant):
    @property
    def n(self): ...

class Int(Num):
    value: int = ...

class Decimal(Num): ...

class Hex(Num):
    @property
    def n_bytes(self): ...

class Str(Constant):
    @property
    def s(self): ...

class Bytes(Constant):
    @property
    def s(self): ...

class NameConstant(Constant): ...
class Ellipsis(Constant): ...

class List(VyperNode):
    elements: list = ...

class Tuple(VyperNode):
    elements: list = ...

class Dict(VyperNode):
    keys: list = ...
    values: list = ...

class Name(ExprNode):
    id: str = ...
    _type: str = ...

class Expr(VyperNode):
    value: ExprNode = ...

class ExtCall(VyperNode):
    value: Call = ...

class StaticCall(VyperNode):
    value: Call = ...

class UnaryOp(ExprNode):
    op: VyperNode = ...
    operand: ExprNode = ...

class USub(VyperNode): ...
class Not(VyperNode): ...

class BinOp(ExprNode):
    op: VyperNode = ...
    left: ExprNode = ...
    right: ExprNode = ...

class Add(VyperNode): ...
class Sub(VyperNode): ...
class Mult(VyperNode): ...
class Div(VyperNode): ...
class FloorDiv(VyperNode): ...
class Mod(VyperNode): ...
class Pow(VyperNode): ...
class LShift(VyperNode): ...
class RShift(VyperNode): ...
class BitAnd(VyperNode): ...
class BitOr(VyperNode): ...
class BitXor(VyperNode): ...

class BoolOp(ExprNode):
    op: VyperNode = ...
    values: list[ExprNode] = ...

class And(VyperNode): ...
class Or(VyperNode): ...

class Compare(ExprNode):
    op: VyperNode = ...
    left: ExprNode = ...
    right: ExprNode = ...

class Eq(VyperNode): ...
class NotEq(VyperNode): ...
class Lt(VyperNode): ...
class LtE(VyperNode): ...
class Gt(VyperNode): ...
class GtE(VyperNode): ...
class In(VyperNode): ...
class NotIn(VyperNode): ...

class Call(ExprNode):
    args: list = ...
    keywords: list = ...
    func: ExprNode = ...
    is_extcall: bool = ...
    is_staticcall: bool = ...
    is_plain_call: bool = ...
    kind_str: str = ...

class keyword(VyperNode): ...

class Attribute(ExprNode):
    attr: str = ...
    value: ExprNode = ...

class Subscript(ExprNode):
    slice: ExprNode = ...
    value: ExprNode = ...

class Assign(VyperNode): ...

class AnnAssign(VyperNode):
    target: Name = ...
    value: VyperNode = ...
    annotation: VyperNode = ...

class VariableDecl(VyperNode):
    target: Name = ...
    value: VyperNode = ...
    annotation: VyperNode = ...
    is_constant: bool = ...
    is_public: bool = ...
    is_immutable: bool = ...
    _expanded_getter: FunctionDef = ...

class AugAssign(VyperNode):
    op: VyperNode = ...
    target: ExprNode = ...
    value: ExprNode = ...

class Raise(VyperNode): ...
class Assert(VyperNode): ...
class Pass(VyperNode): ...

class Import(VyperNode):
    alias: str = ...
    name: str = ...

class ImportFrom(VyperNode):
    alias: str = ...
    level: int = ...
    module: str = ...
    name: str = ...

class ImplementsDecl(VyperNode):
    target: Name = ...
    annotation: Name = ...

class UsesDecl(VyperNode):
    annotation: VyperNode = ...

class InitializesDecl(VyperNode):
    annotation: VyperNode = ...

class ExportsDecl(VyperNode):
    annotation: VyperNode = ...

class If(VyperNode):
    body: list = ...
    orelse: list = ...

class IfExp(ExprNode):
    test: ExprNode = ...
    body: ExprNode = ...
    orelse: ExprNode = ...

class NamedExpr(ExprNode):
    target: Name = ...
    value: ExprNode = ...

class For(VyperNode):
    target: ExprNode
    iter: ExprNode
    body: list[VyperNode]

class Break(VyperNode): ...
class Continue(VyperNode): ...



# Vyper documentation build configuration file, created by
# sphinx-quickstart on Wed Jul 26 11:18:29 2017.

extensions = [
    "sphinx_copybutton",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
master_doc = "toctree"

# General information about the project.
project = "Vyper"
copyright = "2017-2024 CC-BY-4.0 Vyper Team"
author = "Vyper Team (originally created by Vitalik Buterin)"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "vyper"

# -- Options for HTML output ----------------------------------------------
html_theme = "shibuya"
html_theme_options = {
    "accent_color": "purple",
    "twitter_creator": "vyperlang",
    "twitter_site": "vyperlang",
    "twitter_url": "https://twitter.com/vyperlang",
    "github_url": "https://github.com/vyperlang",
}
html_favicon = "logo.svg"
html_logo = "logo.svg"

# For the "Edit this page ->" link
html_context = {
    "source_type": "github",
    "source_user": "vyperlang",
    "source_repo": "vyper",
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Vyperdoc"


# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "Vyper.tex",
        "Vyper Documentation",
        author,
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "vyper", "Vyper Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Vyper",
        "Vyper Documentation",
        author,
        "Vyper",
        "One line description of project.",
        "Miscellaneous",
    ),
]

intersphinx_mapping = {
    "brownie": ("https://eth-brownie.readthedocs.io/en/stable", None),
    "pytest": ("https://docs.pytest.org/en/latest/", None),
    "python": ("https://docs.python.org/3.10/", None),
}



def test_convert_from_bool(get_contract):
    code = """
@external
def foo() -> bool:
    val: bool = True and True and False
    return val

@external
def bar() -> bool:
    val: bool = True or True or False
    return val

@external
def foobar() -> bool:
    val: bool = False and True or False
    return val

@external
def oof() -> bool:
    val: bool = False or False or False or False or False or True
    return val

@external
def rab() -> bool:
    val: bool = True and True and True and True and True and False
    return val

@external
def oofrab() -> bool:
    val: bool = False and True or False and True or False and False or True
    return val
    """

    c = get_contract(code)
    assert c.foo() is False
    assert c.bar() is True
    assert c.foobar() is False
    assert c.oof() is True
    assert c.rab() is False
    assert c.oofrab() is True



# wrapper module around whatever encoder we are using
from decimal import Decimal
from typing import Any

from eth.codecs.abi.decoder import Decoder
from eth.codecs.abi.encoder import Encoder
from eth.codecs.abi.exceptions import ABIError
from eth.codecs.abi.nodes import ABITypeNode, AddressNode, BytesNode, FixedNode
from eth.codecs.abi.parser import Parser
from hexbytes import HexBytes

_parsers: dict[str, ABITypeNode] = {}


class _Encoder(Encoder):
    """
    Custom encoder that converts some types to the expected format.
    """

    @classmethod
    def visit_BytesNode(cls, node: BytesNode, value: bytes | str) -> bytes:
        if isinstance(value, str):
            assert value.startswith("0x"), "Sanity check failed: expected hex string"
            value = bytes.fromhex(value[2:])
        return super().visit_BytesNode(node, value)

    @classmethod
    def visit_FixedNode(cls, node: FixedNode, value: Decimal | int) -> bytes:
        # REVIEW: note coming changes to decimal (PR #3696)
        if isinstance(value, int):
            value = Decimal(value)
        return super().visit_FixedNode(node, value)

    @classmethod
    def visit_AddressNode(cls, node: AddressNode, value: str | bytes | HexBytes) -> bytes:
        if isinstance(value, HexBytes):
            value = value.hex()
        if isinstance(value, bytes):
            value = "0x" + value.hex()
        return super().visit_AddressNode(node, value)


def _get_parser(schema: str):
    try:
        return _parsers[schema]
    except KeyError:
        _parsers[schema] = (ret := Parser.parse(schema))
        return ret


def abi_encode(schema: str, data: Any) -> bytes:
    return _Encoder.encode(_get_parser(schema), data)


def abi_decode(schema: str, data: bytes) -> Any:
    return Decoder.decode(_get_parser(schema), data)


def is_abi_encodable(abi_type: str, data: Any) -> bool:
    try:
        abi_encode(abi_type, data)
        return True
    except ABIError:
        return False



import pytest
from eth_utils import to_wei

from tests.utils import decimal_to_int
from vyper.semantics.types import DecimalT
from vyper.utils import quantize, round_towards_zero

wei_denoms = {
    "femtoether": 3,
    "kwei": 3,
    "babbage": 3,
    "picoether": 6,
    "mwei": 6,
    "lovelace": 6,
    "nanoether": 9,
    "gwei": 9,
    "shannon": 9,
    "microether": 12,
    "szabo": 12,
    "milliether": 15,
    "finney": 15,
    "ether": 18,
    "kether": 21,
    "grand": 21,
}


@pytest.mark.parametrize("denom,multiplier", wei_denoms.items())
def test_wei_uint256(get_contract, tx_failed, denom, multiplier):
    code = f"""
@external
def foo(a: uint256) -> uint256:
    return as_wei_value(a, "{denom}")
    """

    c = get_contract(code)

    value = (2**256 - 1) // (10**multiplier)
    assert c.foo(value) == value * (10**multiplier)

    value = (2**256 - 1) // (10 ** (multiplier - 1))
    with tx_failed():
        c.foo(value)


@pytest.mark.parametrize("denom,multiplier", wei_denoms.items())
def test_wei_int128(get_contract, tx_failed, denom, multiplier):
    code = f"""
@external
def foo(a: int128) -> uint256:
    return as_wei_value(a, "{denom}")
    """

    c = get_contract(code)
    value = (2**127 - 1) // (10**multiplier)

    assert c.foo(value) == value * (10**multiplier)


@pytest.mark.parametrize("denom,multiplier", wei_denoms.items())
def test_wei_decimal(get_contract, tx_failed, denom, multiplier):
    code = f"""
@external
def foo(a: decimal) -> uint256:
    return as_wei_value(a, "{denom}")
    """
    c = get_contract(code)

    denom_int = 10**multiplier
    # TODO: test with more values
    _, hi = DecimalT().ast_bounds
    value = quantize(hi / denom_int)

    assert c.foo(decimal_to_int(value)) == round_towards_zero(value * denom_int)


@pytest.mark.parametrize("value", (-1, -(2**127)))
@pytest.mark.parametrize("data_type", ["decimal", "int128"])
def test_negative_value_reverts(get_contract, tx_failed, value, data_type):
    code = f"""
@external
def foo(a: {data_type}) -> uint256:
    return as_wei_value(a, "ether")
    """

    c = get_contract(code)
    with tx_failed():
        c.foo(value)


@pytest.mark.parametrize("denom,multiplier", wei_denoms.items())
@pytest.mark.parametrize("data_type", ["decimal", "int128", "uint256"])
def test_zero_value(get_contract, tx_failed, denom, multiplier, data_type):
    code = f"""
@external
def foo(a: {data_type}) -> uint256:
    return as_wei_value(a, "{denom}")
    """

    c = get_contract(code)
    assert c.foo(0) == 0


def test_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: uint8) -> uint8: nonpayable

@external
def foo(a: Foo) -> uint256:
    return as_wei_value(extcall a.foo(7), "ether")
    """

    c1 = side_effects_contract("uint8")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == to_wei(7, "ether")
    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_internal_call(get_contract):
    code = """
@external
def foo() -> uint256:
    return as_wei_value(self.bar(), "ether")

@internal
def bar() -> uint8:
    return 7
    """
    c = get_contract(code)
    assert c.foo() == to_wei(7, "ether")



# a compilation unit -- all functions and constructor

from typing import Any, List

import vyper.ast as vy_ast
from vyper.codegen import core, jumptable_utils
from vyper.codegen.core import shr
from vyper.codegen.function_definitions import (
    generate_ir_for_external_function,
    generate_ir_for_internal_function,
)
from vyper.codegen.ir_node import IRnode
from vyper.compiler.settings import _is_debug_mode
from vyper.exceptions import CompilerPanic
from vyper.semantics.types.module import ModuleT
from vyper.utils import OrderedSet, method_id_int


# calculate globally reachable functions to see which
# ones should make it into the final bytecode.
def _runtime_reachable_functions(module_t, id_generator):
    ret = OrderedSet()

    for fn_t in module_t.exposed_functions:
        assert isinstance(fn_t.ast_def, vy_ast.FunctionDef)

        ret.update(fn_t.reachable_internal_functions)
        ret.add(fn_t)

    # create globally unique IDs for each function
    for fn_t in ret:
        id_generator.ensure_id(fn_t)

    return ret


class IDGenerator:
    def __init__(self):
        self._id = 0

    def ensure_id(self, fn_t):
        if fn_t._function_id is None:
            fn_t._function_id = self._id
            self._id += 1


def _is_constructor(func_ast):
    return func_ast._metadata["func_type"].is_constructor


def _is_fallback(func_ast):
    return func_ast._metadata["func_type"].is_fallback


def _is_internal(func_ast):
    return func_ast._metadata["func_type"].is_internal


def _is_payable(func_ast):
    return func_ast._metadata["func_type"].is_payable


def _annotated_method_id(abi_sig):
    method_id = method_id_int(abi_sig)
    annotation = f"{hex(method_id)}: {abi_sig}"
    return IRnode(method_id, annotation=annotation)


def label_for_entry_point(abi_sig, entry_point):
    method_id = method_id_int(abi_sig)
    return f"{entry_point.func_t._ir_info.ir_identifier}{method_id}"


# adapt whatever generate_ir_for_function gives us into an IR node
def _ir_for_fallback_or_ctor(func_ast, *args, **kwargs):
    func_t = func_ast._metadata["func_type"]
    assert func_t.is_fallback or func_t.is_constructor

    ret = ["seq"]
    if not func_t.is_payable:
        callvalue_check = ["assert", ["iszero", "callvalue"]]
        ret.append(IRnode.from_list(callvalue_check, error_msg="nonpayable check"))

    func_ir = generate_ir_for_external_function(func_ast, *args, **kwargs)
    assert len(func_ir.entry_points) == 1

    # add a goto to make the function entry look like other functions
    # (for zksync interpreter)
    ret.append(["goto", func_t._ir_info.external_function_base_entry_label])
    ret.append(func_ir.common_ir)

    return IRnode.from_list(ret)


def _ir_for_internal_function(func_ast, *args, **kwargs):
    return generate_ir_for_internal_function(func_ast, *args, **kwargs).func_ir


def _generate_external_entry_points(external_functions, module_t):
    entry_points = {}  # map from ABI sigs to ir code
    sig_of = {}  # reverse map from method ids to abi sig

    for code in external_functions:
        func_ir = generate_ir_for_external_function(code, module_t)
        for abi_sig, entry_point in func_ir.entry_points.items():
            method_id = method_id_int(abi_sig)
            assert abi_sig not in entry_points
            assert method_id not in sig_of

            entry_points[abi_sig] = entry_point
            sig_of[method_id] = abi_sig

        # stick function common body into final entry point to save a jump
        ir_node = IRnode.from_list(["seq", entry_point.ir_node, func_ir.common_ir])
        entry_point.ir_node = ir_node

    return entry_points, sig_of


# codegen for all runtime functions + callvalue/calldata checks,
# with O(1) jumptable for selector table.
# uses two level strategy: uses `method_id % n_buckets` to descend
# into a bucket (of about 8-10 items), and then uses perfect hash
# to select the final function.
# costs about 212 gas for typical function and 8 bytes of code (+ ~87 bytes of global overhead)
def _selector_section_dense(external_functions, module_t):
    function_irs = []

    if len(external_functions) == 0:
        return IRnode.from_list(["seq"])

    entry_points, sig_of = _generate_external_entry_points(external_functions, module_t)

    # generate the label so the jumptable works
    for abi_sig, entry_point in entry_points.items():
        label = label_for_entry_point(abi_sig, entry_point)
        ir_node = ["label", label, ["var_list"], entry_point.ir_node]
        function_irs.append(IRnode.from_list(ir_node))

    n_buckets, jumptable_info = jumptable_utils.generate_dense_jumptable_info(entry_points.keys())
    # note: we are guaranteed by jumptable_utils that there are no buckets
    # which are empty. sanity check that the bucket ids are well-behaved:
    assert n_buckets == len(jumptable_info)
    for i, (bucket_id, _) in enumerate(sorted(jumptable_info.items())):
        assert i == bucket_id

    #  bucket magic <2 bytes> | bucket location <2 bytes> | bucket size <1 byte>
    # TODO: can make it smaller if the largest bucket magic <= 255
    SZ_BUCKET_HEADER = 5

    selector_section = ["seq"]

    bucket_id = ["mod", "_calldata_method_id", n_buckets]
    bucket_hdr_location = [
        "add",
        ["symbol", "BUCKET_HEADERS"],
        ["mul", bucket_id, SZ_BUCKET_HEADER],
    ]
    # get bucket header
    dst = 32 - SZ_BUCKET_HEADER
    assert dst >= 0

    if _is_debug_mode():
        selector_section.append(["assert", ["eq", "msize", 0]])

    selector_section.append(["codecopy", dst, bucket_hdr_location, SZ_BUCKET_HEADER])

    # figure out the minimum number of bytes we can use to encode
    # min_calldatasize in function info
    largest_mincalldatasize = max(f.min_calldatasize for f in entry_points.values())
    FN_METADATA_BYTES = (largest_mincalldatasize.bit_length() + 7) // 8

    func_info_size = 4 + 2 + FN_METADATA_BYTES
    # grab function info.
    # method id <4 bytes> | label <2 bytes> | func info <1-3 bytes>
    # func info (1-3 bytes, packed) for: expected calldatasize, is_nonpayable bit
    # NOTE: might be able to improve codesize if we use variable # of bytes
    # per bucket

    hdr_info = IRnode.from_list(["mload", 0])
    with hdr_info.cache_when_complex("hdr_info") as (b1, hdr_info):
        bucket_location = ["and", 0xFFFF, shr(8, hdr_info)]
        bucket_magic = shr(24, hdr_info)
        bucket_size = ["and", 0xFF, hdr_info]
        # ((method_id * bucket_magic) >> BITS_MAGIC) % bucket_size
        func_id = [
            "mod",
            shr(jumptable_utils.BITS_MAGIC, ["mul", bucket_magic, "_calldata_method_id"]),
            bucket_size,
        ]
        func_info_location = ["add", bucket_location, ["mul", func_id, func_info_size]]
        dst = 32 - func_info_size
        assert func_info_size >= SZ_BUCKET_HEADER  # otherwise mload will have dirty bytes
        assert dst >= 0
        selector_section.append(b1.resolve(["codecopy", dst, func_info_location, func_info_size]))

    func_info = IRnode.from_list(["mload", 0])
    fn_metadata_mask = 2 ** (FN_METADATA_BYTES * 8) - 1
    calldatasize_mask = fn_metadata_mask - 1  # ex. 0xFFFE
    with func_info.cache_when_complex("func_info") as (b1, func_info):
        x = ["seq"]

        # expected calldatasize always satisfies (x - 4) % 32 == 0
        # the lower 5 bits are always 0b00100, so we can use those
        # bits for other purposes.
        is_nonpayable = ["and", 1, func_info]
        expected_calldatasize = ["and", calldatasize_mask, func_info]

        label_bits_ofst = FN_METADATA_BYTES * 8
        function_label = ["and", 0xFFFF, shr(label_bits_ofst, func_info)]
        method_id_bits_ofst = (FN_METADATA_BYTES + 2) * 8
        function_method_id = shr(method_id_bits_ofst, func_info)

        # check method id is right, if not then fallback.
        # need to check calldatasize >= 4 in case there are
        # trailing 0s in the method id.
        calldatasize_valid = ["gt", "calldatasize", 3]
        method_id_correct = ["eq", function_method_id, "_calldata_method_id"]
        should_fallback = ["iszero", ["and", calldatasize_valid, method_id_correct]]
        x.append(["if", should_fallback, ["goto", "fallback"]])

        # assert callvalue == 0 if nonpayable
        bad_callvalue = ["mul", is_nonpayable, "callvalue"]
        # assert calldatasize at least minimum for the abi type
        bad_calldatasize = ["lt", "calldatasize", expected_calldatasize]
        failed_entry_conditions = ["or", bad_callvalue, bad_calldatasize]
        check_entry_conditions = IRnode.from_list(
            ["assert", ["iszero", failed_entry_conditions]],
            error_msg="bad calldatasize or callvalue",
        )
        x.append(check_entry_conditions)
        jump_targets = [func.args[0].value for func in function_irs]
        jump_instr = IRnode.from_list(["djump", function_label, *jump_targets])
        x.append(jump_instr)
        selector_section.append(b1.resolve(x))

    bucket_headers = ["data", "BUCKET_HEADERS"]

    for bucket_id, bucket in sorted(jumptable_info.items()):
        bucket_headers.append(bucket.magic.to_bytes(2, "big"))
        bucket_headers.append(["symbol", f"bucket_{bucket_id}"])
        # note: buckets are usually ~10 items. to_bytes would
        # fail if the int is too big.
        bucket_headers.append(bucket.bucket_size.to_bytes(1, "big"))

    selector_section.append(bucket_headers)

    for bucket_id, bucket in jumptable_info.items():
        function_infos = ["data", f"bucket_{bucket_id}"]
        # sort function infos by their image.
        for method_id in bucket.method_ids_image_order:
            abi_sig = sig_of[method_id]
            entry_point = entry_points[abi_sig]

            method_id_bytes = method_id.to_bytes(4, "big")
            symbol = ["symbol", label_for_entry_point(abi_sig, entry_point)]
            func_metadata_int = entry_point.min_calldatasize | int(
                not entry_point.func_t.is_payable
            )
            func_metadata = func_metadata_int.to_bytes(FN_METADATA_BYTES, "big")

            function_infos.extend([method_id_bytes, symbol, func_metadata])

        selector_section.append(function_infos)

    ret = ["seq", ["with", "_calldata_method_id", shr(224, ["calldataload", 0]), selector_section]]

    ret.extend(function_irs)

    return ret


# codegen for all runtime functions + callvalue/calldata checks,
# with O(1) jumptable for selector table.
# uses two level strategy: uses `method_id % n_methods` to calculate
# a bucket, and then descends into linear search from there.
# costs about 126 gas for typical (nonpayable, >0 args, avg bucket size 1.5)
# function and 24 bytes of code (+ ~23 bytes of global overhead)
def _selector_section_sparse(external_functions, module_t):
    ret = ["seq"]

    if len(external_functions) == 0:
        return ret

    entry_points, sig_of = _generate_external_entry_points(external_functions, module_t)

    n_buckets, buckets = jumptable_utils.generate_sparse_jumptable_buckets(entry_points.keys())

    # 2 bytes for bucket location
    SZ_BUCKET_HEADER = 2

    if n_buckets > 1:
        bucket_id = ["mod", "_calldata_method_id", n_buckets]
        bucket_hdr_location = [
            "add",
            ["symbol", "selector_buckets"],
            ["mul", bucket_id, SZ_BUCKET_HEADER],
        ]
        # get bucket header
        dst = 32 - SZ_BUCKET_HEADER
        assert dst >= 0

        if _is_debug_mode():
            ret.append(["assert", ["eq", "msize", 0]])

        ret.append(["codecopy", dst, bucket_hdr_location, SZ_BUCKET_HEADER])

        jump_targets = []

        for i in range(n_buckets):
            if i in buckets:
                bucket_label = f"selector_bucket_{i}"
                jump_targets.append(bucket_label)
            else:
                # empty bucket
                jump_targets.append("fallback")

        jumptable_data = ["data", "selector_buckets"]
        jumptable_data.extend(["symbol", label] for label in jump_targets)

        jumpdest = IRnode.from_list(["mload", 0])

        jump_instr = IRnode.from_list(["djump", jumpdest, *jump_targets])
        ret.append(jump_instr)
        ret.append(jumptable_data)

    for bucket_id, bucket in buckets.items():
        bucket_label = f"selector_bucket_{bucket_id}"
        ret.append(["label", bucket_label, ["var_list"], ["seq"]])

        handle_bucket = ["seq"]

        for method_id in bucket:
            sig = sig_of[method_id]
            entry_point = entry_points[sig]
            func_t = entry_point.func_t
            expected_calldatasize = entry_point.min_calldatasize

            dispatch = ["seq"]  # code to dispatch into the function
            skip_callvalue_check = func_t.is_payable
            skip_calldatasize_check = expected_calldatasize == 4
            bad_callvalue = [0] if skip_callvalue_check else ["callvalue"]
            bad_calldatasize = (
                [0] if skip_calldatasize_check else ["lt", "calldatasize", expected_calldatasize]
            )

            dispatch.append(
                IRnode.from_list(
                    ["assert", ["iszero", ["or", bad_callvalue, bad_calldatasize]]],
                    error_msg="bad calldatasize or callvalue",
                )
            )
            # we could skip a jumpdest per method if we out-lined the entry point
            # so the dispatcher looks just like -
            # ```(if (eq <calldata_method_id> method_id)
            #   (goto entry_point_label))```
            # it would another optimization for patterns like
            # `if ... (goto)` though.
            dispatch.append(entry_point.ir_node)

            method_id_check = ["eq", "_calldata_method_id", _annotated_method_id(sig)]
            has_trailing_zeroes = method_id.to_bytes(4, "big").endswith(b"\x00")
            if has_trailing_zeroes:
                # if the method id check has trailing 0s, we need to include
                # a calldatasize check to distinguish from when not enough
                # bytes are provided for the method id in calldata.
                method_id_check = ["and", ["ge", "calldatasize", 4], method_id_check]
            handle_bucket.append(["if", method_id_check, dispatch])

        # close out the bucket with a goto fallback so we don't keep searching
        handle_bucket.append(["goto", "fallback"])

        ret.append(handle_bucket)

    ret = ["seq", ["with", "_calldata_method_id", shr(224, ["calldataload", 0]), ret]]

    return ret


# codegen for all runtime functions + callvalue/calldata checks,
# O(n) linear search for the method id
# mainly keep this in for backends which cannot handle the indirect jump
# in selector_section_dense and selector_section_sparse
def _selector_section_linear(external_functions, module_t):
    ret = ["seq"]
    if len(external_functions) == 0:
        return ret

    ret.append(["if", ["lt", "calldatasize", 4], ["goto", "fallback"]])

    entry_points, sig_of = _generate_external_entry_points(external_functions, module_t)

    dispatcher = ["seq"]

    for sig, entry_point in entry_points.items():
        func_t = entry_point.func_t
        expected_calldatasize = entry_point.min_calldatasize

        dispatch = ["seq"]  # code to dispatch into the function

        if not func_t.is_payable:
            callvalue_check = ["assert", ["iszero", "callvalue"]]
            dispatch.append(IRnode.from_list(callvalue_check, error_msg="nonpayable check"))

        good_calldatasize = ["ge", "calldatasize", expected_calldatasize]
        calldatasize_check = ["assert", good_calldatasize]
        dispatch.append(IRnode.from_list(calldatasize_check, error_msg="calldatasize check"))

        dispatch.append(entry_point.ir_node)

        method_id_check = ["eq", "_calldata_method_id", _annotated_method_id(sig)]
        dispatcher.append(["if", method_id_check, dispatch])

    ret.append(["with", "_calldata_method_id", shr(224, ["calldataload", 0]), dispatcher])

    return ret


# take a ModuleT, and generate the runtime and deploy IR
def generate_ir_for_module(module_t: ModuleT) -> tuple[IRnode, IRnode]:
    # order functions so that each function comes after all of its callees
    id_generator = IDGenerator()
    runtime_reachable = _runtime_reachable_functions(module_t, id_generator)

    function_defs = [fn_t.ast_def for fn_t in runtime_reachable]

    runtime_functions = [f for f in function_defs if not _is_constructor(f)]
    internal_functions = [f for f in runtime_functions if _is_internal(f)]
    external_functions = [
        f for f in runtime_functions if not _is_internal(f) and not _is_fallback(f)
    ]
    default_function = next((f for f in runtime_functions if _is_fallback(f)), None)

    internal_functions_ir: list[IRnode] = []

    # module_tinternal functions first so we have the function info
    for func_ast in internal_functions:
        func_ir = _ir_for_internal_function(func_ast, module_t, False)
        internal_functions_ir.append(IRnode.from_list(func_ir))

    if core._opt_none():
        selector_section = _selector_section_linear(external_functions, module_t)
    # dense vs sparse global overhead is amortized after about 4 methods.
    # (--debug will force dense selector table anyway if _opt_codesize is selected.)
    elif core._opt_codesize() and (len(external_functions) > 4 or _is_debug_mode()):
        selector_section = _selector_section_dense(external_functions, module_t)
    else:
        selector_section = _selector_section_sparse(external_functions, module_t)

    if default_function:
        fallback_ir = _ir_for_fallback_or_ctor(default_function, module_t)
    else:
        fallback_ir = IRnode.from_list(
            ["revert", 0, 0], annotation="Default function", error_msg="fallback function"
        )

    runtime = ["seq", selector_section]
    runtime.append(["goto", "fallback"])
    runtime.append(["label", "fallback", ["var_list"], fallback_ir])

    runtime.extend(internal_functions_ir)

    deploy_code: List[Any] = ["seq"]
    immutables_len = module_t.immutable_section_bytes

    if (init_func_t := module_t.init_function) is not None:
        # cleanly rerun codegen for internal functions with `is_ctor_ctx=True`
        id_generator.ensure_id(init_func_t)
        ctor_internal_func_irs = []

        reachable_from_ctor = init_func_t.reachable_internal_functions
        for func_t in reachable_from_ctor:
            id_generator.ensure_id(func_t)
            fn_ast = func_t.ast_def
            func_ir = _ir_for_internal_function(fn_ast, module_t, is_ctor_context=True)
            ctor_internal_func_irs.append(func_ir)

        # generate init_func_ir after callees to ensure they have analyzed
        # memory usage.
        # TODO might be cleaner to separate this into an _init_ir helper func
        init_func_ir = _ir_for_fallback_or_ctor(init_func_t.ast_def, module_t)

        # pass the amount of memory allocated for the init function
        # so that deployment does not clobber while preparing immutables
        # note: (deploy mem_ofst, code, extra_padding)
        init_mem_used = init_func_t._ir_info.frame_info.mem_used

        # force msize to be initialized past the end of immutables section
        # so that builtins which use `msize` for "dynamic" memory
        # allocation do not clobber uninitialized immutables.
        # cf. GH issue 3101.
        # note mload/iload X touches bytes from X to X+32, and msize rounds up
        # to the nearest 32, so `iload`ing `immutables_len - 32` guarantees
        # that `msize` will refer to a memory location of at least
        # `<immutables_start> + immutables_len` (where <immutables_start> ==
        # `_mem_deploy_end` as defined in the assembler).
        # note:
        #   mload 32 => msize == 64
        #   mload 33 => msize == 96
        # assumption in general: (mload X) => msize == ceil32(X + 32)
        # see py-evm extend_memory: after_size = ceil32(start_position + size)
        if immutables_len > 0:
            deploy_code.append(["iload", max(0, immutables_len - 32)])

        deploy_code.append(init_func_ir)
        deploy_code.append(["deploy", init_mem_used, runtime, immutables_len])
        # internal functions come at end of initcode
        deploy_code.extend(ctor_internal_func_irs)

    else:
        if immutables_len != 0:  # pragma: nocover
            raise CompilerPanic("unreachable")
        deploy_code.append(["deploy", 0, runtime, 0])

    # compile all remaining internal functions so that _ir_info is populated
    # (whether or not it makes it into the final IR artifact)
    to_visit: OrderedSet = OrderedSet()
    for func_ast in module_t.function_defs:
        fn_t = func_ast._metadata["func_type"]
        if fn_t.is_internal:
            to_visit.update(fn_t.reachable_internal_functions)
            to_visit.add(fn_t)

    for fn_t in to_visit:
        if fn_t._ir_info is None:
            id_generator.ensure_id(fn_t)
            _ = _ir_for_internal_function(fn_t.ast_def, module_t, False)

    return IRnode.from_list(deploy_code), IRnode.from_list(runtime)



from vyper.exceptions import FunctionDeclarationException


def test_invalid_if_both_public_and_internal(assert_compile_failed, get_contract):
    code = """
@external
@internal
def foo():
    x: uint256 = 1
"""

    assert_compile_failed(lambda: get_contract(code), FunctionDeclarationException)


def test_invalid_if_visibility_isnt_declared(assert_compile_failed, get_contract):
    code = """
def foo():
    x: uint256 = 1
"""

    assert_compile_failed(lambda: get_contract(code), FunctionDeclarationException)



def test_bytes_logging_extended(get_contract, get_logs):
    code = """
event MyLog:
    arg1: int128
    arg2: Bytes[64]
    arg3: int128

@external
def foo():
    log MyLog(667788, b'hellohellohellohellohellohellohellohellohello', 334455)
    """

    c = get_contract(code)
    c.foo()
    (log,) = get_logs(c, "MyLog")

    assert log.args.arg1 == 667788
    assert log.args.arg2 == b"hello" * 9
    assert log.args.arg3 == 334455


def test_bytes_logging_extended_variables(get_contract, get_logs):
    code = """
event MyLog:
    arg1: Bytes[64]
    arg2: Bytes[64]
    arg3: Bytes[64]

@external
def foo():
    a: Bytes[64] = b'hellohellohellohellohellohellohellohellohello'
    b: Bytes[64] = b'hellohellohellohellohellohellohellohello'
    # test literal much smaller than buffer
    log MyLog(a, b, b'hello')
    """

    c = get_contract(code)
    c.foo()
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == b"hello" * 9
    assert log.args.arg2 == b"hello" * 8
    assert log.args.arg3 == b"hello" * 1


def test_bytes_logging_extended_passthrough(get_contract, get_logs):
    code = """
event MyLog:
    arg1: int128
    arg2: Bytes[64]
    arg3: int128

@external
def foo(a: int128, b: Bytes[64], c: int128):
    log MyLog(a, b, c)
    """

    c = get_contract(code)

    c.foo(333, b"flower" * 8, 444)
    log = get_logs(c, "MyLog")

    assert log[0].args.arg1 == 333
    assert log[0].args.arg2 == b"flower" * 8
    assert log[0].args.arg3 == 444


def test_bytes_logging_extended_storage(get_contract, get_logs):
    code = """
event MyLog:
    arg1: int128
    arg2: Bytes[64]
    arg3: int128

a: int128
b: Bytes[64]
c: int128

@external
def foo():
    log MyLog(self.a, self.b, self.c)

@external
def set(x: int128, y: Bytes[64], z: int128):
    self.a = x
    self.b = y
    self.c = z
    """

    c = get_contract(code)
    c.foo()
    log = get_logs(c, "MyLog")

    assert log[0].args.arg1 == 0
    assert log[0].args.arg2 == b""
    assert log[0].args.arg3 == 0

    c.set(333, b"flower" * 8, 444)
    c.foo()

    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == 333
    assert log.args.arg2 == b"flower" * 8
    assert log.args.arg3 == 444


def test_bytes_logging_extended_mixed_with_lists(get_contract, get_logs):
    code = """
event MyLog:
    arg1: int128[2][2]
    arg2: Bytes[64]
    arg3: int128
    arg4: Bytes[64]

@external
def foo():
    log MyLog(
        [[24, 26], [12, 10]],
        b'hellohellohellohellohellohellohellohellohello',
        314159,
        b'helphelphelphelphelphelphelphelphelphelphelp'
    )
    """

    c = get_contract(code)
    c.foo()
    (log,) = get_logs(c, "MyLog")

    assert log.args.arg1 == [[24, 26], [12, 10]]
    assert log.args.arg2 == b"hello" * 9
    assert log.args.arg3 == 314159
    assert log.args.arg4 == b"help" * 11



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    y: int256 = abs(
        -57896044618658097711785492504343953926634992332820282019728792003956564819968
    )
    """,
        TypeMismatch,
    )
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_abs_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
FOO: constant(int256) = -3
BAR: constant(int256) = abs(FOO)

@external
def foo():
    a: int256 = BAR
    """
]


@pytest.mark.parametrize("code", valid_list)
def test_abs_pass(code):
    assert compile_code(code) is not None



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import ImmutableViolation, StateAccessViolation


@pytest.mark.parametrize(
    "bad_code",
    [
        """
x: int128
@external
@view
def foo() -> int128:
    self.x = 5
    return 1""",
        """
@external
@view
def foo() -> int128:
    send(0x1234567890123456789012345678901234567890, 5)
    return 1""",
        """
@external
@view
def foo():
    selfdestruct(0x1234567890123456789012345678901234567890)""",
        """
x: int128
y: int128
@external
@view
def foo() -> int128:
    self.y = 9
    return 5""",
        """
@external
@view
def foo() -> int128:
    x: Bytes[4] = raw_call(
        0x1234567890123456789012345678901234567890, b"cow", max_outsize=4, gas=595757, value=9
    )
    return 5""",
        """
@external
@view
def foo() -> int128:
    x: address = create_minimal_proxy_to(0x1234567890123456789012345678901234567890, value=9)
    return 5""",
        # test constancy in range expressions
        """
glob: int128
@internal
def foo() -> int128:
    self.glob += 1
    return 5
@external
def bar():
    for i: int128 in range(self.foo(), bound=100):
        pass""",
        """
glob: int128
@internal
def foo() -> int128:
    self.glob += 1
    return 5
@external
def bar():
    for i: int128 in [1,2,3,4,self.foo()]:
        pass""",
        """
f:int128

@internal
def a (x:int128):
    self.f = 100

@view
@external
def b():
    self.a(10)""",
        """
interface A:
    def bar() -> uint16: view

@external
@pure
def test(to: address):
    a: A = A(to)
    x: uint16 = staticcall a.bar()
    """,
        """
interface A:
    def bar() -> uint16: nonpayable

@external
@view
def test(to: address):
    a: A = A(to)
    x: uint16 = extcall a.bar()
    """,
        """
interface A:
    def bar() -> uint16: nonpayable

@external
@view
def test(to: address):
    a: A = A(to)
    extcall a.bar()
    """,
        """
a:DynArray[uint16,3]
@deploy
def __init__():
    self.a = [1,2,3]
@view
@external
def bar()->DynArray[uint16,3]:
    x:uint16 = self.a.pop()
    return self.a # return [1,2]
    """,
        """
from ethereum.ercs import IERC20

token: IERC20

@external
@view
def topup(amount: uint256):
    assert extcall self.token.transferFrom(msg.sender, self, amount)
    """,
    ],
)
def test_statefulness_violations(bad_code):
    with raises(StateAccessViolation):
        compiler.compile_code(bad_code)


@pytest.mark.parametrize(
    "bad_code",
    [
        """
@external
def foo(x: int128):
    x = 5
        """,
        """
@external
def test(a: uint256[4]):
    a[0] = 1
        """,
        """
@external
def test(a: uint256[4][4]):
    a[0][1] = 1
        """,
        """
struct Foo:
    a: DynArray[DynArray[uint256, 2], 2]

@external
def foo(f: Foo) -> Foo:
    f.a[1] = [0, 1]
    return f
        """,
    ],
)
def test_immutability_violations(bad_code):
    with raises(ImmutableViolation):
        compiler.compile_code(bad_code)
